import re
from typing import Generator
import tempfile
import filecmp
import os
import sys
import shlex
from rich import get_console
from rich.text import Text
from curvpyutils.shellutils import print_delta
from .util import StrListWithMostCommon

patterns = {
    "start_subst": re.compile(r"@subst\[`(.*)`\]"),
    "end_subst": re.compile(r"@endsubst"),
    "indent": re.compile(r"^(\s*)")
}

class ReplacementCmd:
    def __init__(self, cmd: str, extra_args: list[str], extra_args_leading: list[str]):
        # Expand environment variables before splitting into command parts
        cmd = os.path.expandvars(cmd)
        try:
            cmd_parts = shlex.split(cmd, posix=True)
        except ValueError:
            cmd_parts = cmd.split()
        self.cmd = cmd_parts[0] if len(cmd_parts) > 0 else ""
        self.args = cmd_parts[1:] if len(cmd_parts) > 1 else []
        if extra_args_leading is not None and len(extra_args_leading) > 0:
            self.args.extend(extra_args_leading)
        if extra_args is not None and len(extra_args) > 0:
            self.args.extend(extra_args)
            
    def to_text(self) -> Text:
        return Text.assemble(
            (self.cmd, "bold yellow"),
            (f" {(' '.join(self.args))}", "bold yellow"),
        )

    def __str__(self):
        return self.to_text().__str__()
    
    def __repr__(self):
        return f"ReplacementCmd({self.__str__()})"

    def invoke(self) -> list[str]:
        """
        Invokes the replacement command with the given arguments.
        """    
        import subprocess
        res = subprocess.check_output([self.cmd, *self.args]).decode("utf-8")
        res_lines = res.split("\n")
        if res_lines[-1].strip() == "":
            res_lines = res_lines[:-1]
        ret_arr = [f"{line}\n" for line in res_lines]
        return ret_arr

class SubstBlock:
    def __init__(self, start_line_num: int, end_line_num: int, replacement_cmd: ReplacementCmd):
        self.start_line_num = start_line_num
        self.end_line_num = end_line_num
        self.replacement_cmd = replacement_cmd
    

    def to_text(self) -> Text:
        t = Text.assemble(
            ("SubstBlock", "purple"),
            ("(from line ", ""),
            (str(self.start_line_num), "bold yellow"),
            (" to ", ""),
            (str(self.end_line_num), "bold yellow"),
            (")", ""),
            (" with replacement command [", ""),
            self.replacement_cmd.to_text(),
            ("]", ""),
        )
        return t

    def __str__(self):
        return f"SubstBlock(from line {self.start_line_num} to {self.end_line_num} with replacement command `{str(self.replacement_cmd)}`)"
    
    def invoke_replacement_cmd(self) -> list[str]:
        return self.replacement_cmd.invoke()

def find_subst_blocks_iter(file_path: str, extra_args: list[str], extra_args_leading: list[str]) -> Generator[SubstBlock, None, None]:
    with open(file_path, "r") as file:
        lines = file.readlines()
    in_subst_block = False
    start_line_num = None
    end_line_num = None
    for line_num, line in enumerate(lines):
        # Use re.search to find the pattern anywhere in the line.
        match = patterns["start_subst"].search(line)
        if match:
            in_subst_block = True
            start_line_num = line_num
            replacement_cmd = ReplacementCmd(match.group(1), extra_args=extra_args, extra_args_leading=extra_args_leading)
        if in_subst_block:
            # Use re.search to find the pattern anywhere in the line.
            match = patterns["end_subst"].search(line)
            if match:
                in_subst_block = False
                end_line_num = line_num
                yield SubstBlock(start_line_num, end_line_num, replacement_cmd)

def get_most_common_indent_str(lines_being_replaced: list[str]) -> str:
    """
    Returns a StrListWithMostCommon of the indent strings in the lines being replaced.
    """
    indent_strs:StrListWithMostCommon = StrListWithMostCommon()
    for line in lines_being_replaced:
        try:
            indent_strs.append(patterns["indent"].match(line).group(1))
        except:
            pass
    return str(indent_strs)

def replace_subst_block(filename: str, output_filename: str, verbose: int = 0, extra_args: list[str] = [], extra_args_leading: list[str] = []):
    """
    Replaces the subst block with the replacement lines.
    """
    if verbose > 1:
        print(f"replace_subst_block: filename={filename}, output_filename={output_filename}")
    with open(filename, "r") as file:
        lines = file.readlines()
    output_lines = []
    last_line_written_out = 0
    for subst_block in find_subst_blocks_iter(filename, extra_args=extra_args, extra_args_leading=extra_args_leading):
        if verbose > 1:
            get_console().print("replace_subst_block: found subst block ", subst_block.to_text())
        replacement_lines = subst_block.invoke_replacement_cmd()
        indentation_str = get_most_common_indent_str(lines[subst_block.start_line_num:subst_block.end_line_num + 1])
        for line in lines[last_line_written_out:subst_block.start_line_num+1]:
            output_lines.append(line)
        for line in replacement_lines:
            output_lines.append(indentation_str + line)
        output_lines.append(lines[subst_block.end_line_num])
        last_line_written_out = subst_block.end_line_num+1
    for line in lines[last_line_written_out:]:
        output_lines.append(line)
    with open(output_filename, "w") as file:
        file.writelines(output_lines)

def run_substitution_on_file(file_path: str, dry_run: bool = False, verbose: int = 0, force: bool = False, disable_colors: bool = False, extra_args: list[str] = [], extra_args_leading: list[str] = [], debug_output_dir: str|None = None) -> bool:
    """
    Runs substitution on a single file, with prompting and a preview diff.

    Returns True if the file was modified, False otherwise.
    """

    file_modified = False

    try:
        temp_out_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        if verbose > 1:
            get_console().print(f"run_substitution_on_file: file_path=[yellow]{file_path}[/yellow], temp_out_file.name=[yellow]{temp_out_file.name}[/yellow]")
        replace_subst_block(file_path, temp_out_file.name, verbose=verbose, extra_args=extra_args, extra_args_leading=extra_args_leading)
        temp_out_file.close()

        # Compare files byte by byte to see if they differ
        if filecmp.cmp(file_path, temp_out_file.name, shallow=False):
            if verbose > 1:
                get_console().print(f"no changes needed for [yellow]{file_path}[/yellow]")
        else:
            if dry_run or not force:
                # show the delta
                print_delta(file_path, temp_out_file.name)

            # prompt for confirmation before overwriting
            if not dry_run:
                if not force:
                    response = input("Overwrite original file with changes? [y/N] ")
                    if response.lower() != 'y':
                        get_console().print(f"aborting - no changes made for [yellow]{file_path}[/yellow]")
                        return False
                if debug_output_dir is not None:
                    os.makedirs(debug_output_dir, exist_ok=True)
                    os.rename(temp_out_file.name, os.path.join(debug_output_dir, os.path.basename(file_path)))
                else:
                    os.rename(temp_out_file.name, file_path)
                file_modified = True
            else:
                # if it is a dry run, don't overwrite anything
                if debug_output_dir is not None:
                    os.makedirs(debug_output_dir, exist_ok=True)
                    os.rename(temp_out_file.name, os.path.join(debug_output_dir, os.path.basename(file_path)))
                    if verbose > 1:
                        get_console().print(f"replace_subst_block: dry run, not overwriting file {file_path}, but writing to {os.path.join(debug_output_dir, os.path.basename(file_path))}")
                else:
                    if verbose > 1:
                        get_console().print(f"replace_subst_block: dry run, not overwriting file {file_path}")
    except Exception as e:
        get_console().print(f"error (at line {sys.exc_info()[2].tb_lineno}): {e}")
        raise e
    finally:
        if os.path.exists(temp_out_file.name):
            os.unlink(temp_out_file.name)
    return file_modified



import os
import sys
import subprocess
from pathlib import Path


def _compute_repo_root() -> str:
    try:
        cp = subprocess.run(["git", "rev-parse", "--show-toplevel"], capture_output=True, text=True, check=True)
        return cp.stdout.strip()
    except Exception:
        # Fallback: walk up to nearest .git
        p = Path(__file__).resolve()
        for parent in p.parents:
            if (parent / ".git").exists():
                return str(parent)
        return str(Path(__file__).resolve().parents[5])


def pytest_sessionstart(session):
    # Ensure sibling helper is importable from subdirs
    this_dir = Path(__file__).resolve().parent
    if str(this_dir) not in sys.path:
        sys.path.insert(0, str(this_dir))

    REL_FAKE_CURV_ROOT = "packages/curvtools/test/curvtools/curvcfg/fake_curv_root"

    repo_root = _compute_repo_root()
    fake_root = os.path.join(repo_root, REL_FAKE_CURV_ROOT)
    os.environ.setdefault("CURV_ROOT_DIR", fake_root)
    os.environ.setdefault("CURV_FAKE_ROOT_REL", REL_FAKE_CURV_ROOT)



import click
from importlib import metadata as ilmd
from curvtools import get_curvtools_version_str


def _dist_name_for_this_package() -> str:
    top = __package__.split('.')[0]
    return ilmd.packages_distributions().get(top, [top])[0]


def get_urls() -> dict[str, str]:
    urls: dict[str, str] = {}
    _DIST = _dist_name_for_this_package()
    try:
        meta = ilmd.metadata(_DIST)
    except Exception:
        return urls
    for key, value in meta.items():
        if key.startswith("Project-URL"):
            url_key, url_value = value.split(",", 1)
            urls[url_key.strip()] = url_value.strip()
    return urls

def get_styled_version(prepend_v: bool = False) -> str:
    """
    Return styled version for click.version_option.
    """
    version = get_curvtools_version_str(short_version=True)
    version_styled = click.style(f"{'v' if prepend_v else ''}{version}", fg="bright_yellow")
    return version_styled

def strip_github_url(url: str) -> str:
    """
    Strip the https prefix and any suffix from a GitHub URL.
    """
    url = url.replace("https://", "")
    url = url.replace(".git", "")
    url = url.rstrip("/")
    return url

def get_styled_version_message() -> str:
    """
    Return styled version message for click.version_option.
    Based on URLs in pyproject.toml's [project.urls] section.
    """
    version_styled = f" {get_styled_version(prepend_v=True)} "

    # get styled program name and version
    program_name_styled = click.style("curvcfg", fg="bright_yellow")

    # get styled github URL
    github_url = get_urls().get("repository", "https://github.com/curvcpu/curv")
    github_url_styled = click.style(strip_github_url(github_url), fg="cyan")
    
    message = program_name_styled + \
        version_styled + \
        f"- Curv RISC-V CPU build config tool (" + \
        github_url_styled + \
        ")"
    return message



import argparse
import os
from pathlib import Path
import sys
from typing import Any, Dict
from dotenv import dotenv_values
import re

CACHE_README_FILENAME = "cache_readme.txt"

_BOOL_TRUE  = {"1", "true", "yes", "on", "y", "t"}
_BOOL_FALSE = {"0", "false", "no", "off", "n", "f"}
_INT_RE = re.compile(r'^[\+\-]?(0[xX][0-9a-fA-F_]+|\d[\d_]*)$')

class DefaultValue():
    def __init__(self, default_value: Any, coerce_to_type: type):
        self.coerce_to_type = coerce_to_type
        self.default_value = self.coerce_value(default_value)
    def coerce_value(self, value: Any) -> Any:
        if value is None:
            return None
        value_str = str(value).strip()
        if self.coerce_to_type is int:
            if _INT_RE.match(value_str):
                value_str = value_str.replace("_", "")
                # Avoid int("08", 0) -> ValueError; prefer base 10 unless explicit hex
                base = 0 if value_str.lower().startswith(("+0x", "-0x", "0x")) else 10
                try:
                    return int(value_str, base)
                except ValueError:
                    return value_str  # not an int-looking string; leave it
        elif self.coerce_to_type is bool:
            low = value_str.lower()
            if low in _BOOL_TRUE:
                return True
            if low in _BOOL_FALSE:
                return False
            raise ValueError(f"Invalid boolean literal for env key: '{value_str!r}'")
        elif self.coerce_to_type is type(None):
            if value_str.lower() in {"none"} or value_str == "":
                return None
            return value_str
        elif self.coerce_to_type is str:
            return value_str
        else:
            return value

DEFAULT_ENV_CFG: Dict[str, DefaultValue] = {
    'CFG_CACHE_SETS': DefaultValue(4, int),
    'CFG_CACHE_WAYS': DefaultValue(2, int),
    'CFG_CACHE_CACHEABLE_ADDRESSES_WIDTH': DefaultValue(27, int),
    'CFG_CACHE_CACHELINES_NUM_WORDS': DefaultValue(16, int),
    'CFG_CACHE_CACHELINES_LATENCY': DefaultValue(0, int),
    'CFG_CACHE_CACHELINES_INITIALLY_VALID': DefaultValue(1, bool),
    'CFG_CACHE_TAGS_HAVE_VALID_DIRTY_BITS': DefaultValue(0, bool),
    'CFG_CACHE_HEX_FILES_BASE_ADDR': DefaultValue(0, int),
    'CFG_CACHE_HEX_FILES_SUBDIRS_ICACHE': DefaultValue(None, str),
    'CFG_CACHE_HEX_FILES_SUBDIRS_DCACHE': DefaultValue(None, str),
    'CFG_CACHE_HEX_FILES_SUBDIRS_CACHELINES': DefaultValue(None, str),
    'CFG_CACHE_HEX_FILES_SUBDIRS_TAGRAM': DefaultValue(None, str),
}

def load_env_into_cfg(path: str) -> Dict[str, Any]:
    """
    Load KEY=VALUE pairs from a .env file using python-dotenv, then coerce
    booleans/ints/None where reasonable. Raises if file missing.
    """
    p = Path(path)
    if not p.is_file():
        raise FileNotFoundError(f".env file not found: {path}")
    # dotenv handles whitespace, comments, quotes, escapes, interpolation, etc.
    raw = dotenv_values(dotenv_path=p, interpolate=True, encoding="utf-8")

    # raw: dict[str, Optional[str]]
    coerced: Dict[str, Any] = {}
    for k, v in raw.items():
        if k in DEFAULT_ENV_CFG:
            coerced[k] = DEFAULT_ENV_CFG[k].coerce_value(v)
        else:
            coerced[k] = v
    return coerced

def get_defaults(env_file_cfg: Dict[str, Any] | None) -> Dict[str, Any]:
    """
    Get defaults, deferring to environment variables (if provided) with lower precedence than env_file_cfg (if provided)
    """

    ret: Dict[str, Any] = {k: DEFAULT_ENV_CFG[k].default_value for k in DEFAULT_ENV_CFG.keys()}

    # environment variables take precedence over defaults
    for k, v in os.environ.items():
        if k in DEFAULT_ENV_CFG:
            ret[k] = DEFAULT_ENV_CFG[k].coerce_value(v)

    # .env file values take precedence over defaults and environment variables
    if env_file_cfg is not None:
        for k, v in env_file_cfg.items():
            if k in DEFAULT_ENV_CFG:
                ret[k] = v
    return ret

def _preparse_env(argv):
    # No -h here; this parser only peeks -e and later becomes a parent for help text
    pre = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
    pre.add_argument(
        "-e", "--config-env-file",
        dest="config_env_file",
        metavar="FILE",
        help="Path to .env with defaults (KEY=VALUE)."
    )
    ns, _ = pre.parse_known_args(argv)
    return ns.config_env_file, pre

def auto_int(x: str) -> int:
    # decimal or 0x... hex
    return int(x, 0)

def build_main_parser(defaults: Dict[str, Any] | None, parent: argparse.ArgumentParser) -> argparse.ArgumentParser:
    # build parser
    parser = argparse.ArgumentParser(
        description="Configurable cache tool for 2-way set associative cache",
        parents=[parent],  # makes -e show up in --help
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument("HEX_FILE", help="Input hex file (32-bit words per line).")
    parser.add_argument("-o", "--output-dir", required=True, help="Output directory.")
    # NOTE: -e comes from the parent; no need to redefine it here.

    cache_config_group = parser.add_argument_group("Cache Configuration")
    cache_config_group.add_argument("--num-sets", "-s", type=int, choices=[4, 8, 16, 32, 64, 128],
                   default=defaults['CFG_CACHE_SETS'],
                   help="Number of sets in the cache, must be a power of 2 between 4 and 128")
    cache_config_group.add_argument("--num-ways", "-w", type=int, choices=[2],
                   default=defaults['CFG_CACHE_WAYS'],
                   help="Number of ways in the cache, currently only 2-way associative is supported")
    cache_config_group.add_argument("--address-width", "-a", type=int,
                   default=defaults['CFG_CACHE_CACHEABLE_ADDRESSES_WIDTH'],
                   help="Width of the highest cacheable address (in bits), must be between 10 and 32")
    cache_config_group.add_argument("--tags-have-valid-dirty-bits",
                   action=argparse.BooleanOptionalAction,
                   default=bool(defaults['CFG_CACHE_TAGS_HAVE_VALID_DIRTY_BITS']),
                   help="Whether tags have two low bits appended: valid, dirty")
    cache_config_group.add_argument("--base-address", "-b", type=auto_int,
                   default=defaults['CFG_CACHE_HEX_FILES_BASE_ADDR'],
                   help="Base address, in bytes (decimal or 0x...). For combined mode, this is the base address of the I$.  For icache-only or dcache-only mode, this is the base address of the I$ or D$ respectively.")
    cache_config_group.add_argument("--cachelines-num-words", "-W", type=int, choices=[16],
                   default=defaults['CFG_CACHE_CACHELINES_NUM_WORDS'],
                   help="Number of 32-bit words per cache line")
    cache_config_group.add_argument("--cachelines-latency", "-l", type=int, choices=[0, 1, 2],
                   default=defaults['CFG_CACHE_CACHELINES_LATENCY'],
                   help="Latency of the cache line, 0: no latency, 1: 1 cycle latency, 2: 2 cycle latency; this affects the type of ram used and therefore fmax")
    cache_config_group.add_argument("--cachelines-initially-valid",
                   action=argparse.BooleanOptionalAction,
                   default=bool(defaults['CFG_CACHE_CACHELINES_INITIALLY_VALID']),
                   help="Whether the cache line is initially valid; set to true if you are prepopulating cache lines from a hex file, false if they will be filled with garbage or zeros")

    output_gen_group = parser.add_argument_group("Output Generation")
    which_mode = output_gen_group.add_mutually_exclusive_group()
    which_mode.add_argument(
        "--mode",
        choices=["icache", "dcache", "combined"],
        default="combined",
        help="which caches to generate: icache means generate I$ only and interpret HEX_FILE input file as the words of I$ only; dcache means the same for D$, and combined means generate both and interpret HEX_FILE as I$ concatenated with D$ (first half of file is I$, second half is D$)"
    )
    # alises for modes enum
    which_mode.add_argument("--icache-only", "-i", dest="mode", action="store_const", const="icache",
                            help="alias for --mode=icache; generate I$ only and interpret HEX_FILE input file as the words of I$ only")
    which_mode.add_argument("--dcache-only", "-d", dest="mode", action="store_const", const="dcache",
                            help="alias for --mode=dcache; generate D$ only and interpret HEX_FILE input file as the words of D$ only")
    which_mode.add_argument("--combined-icache-dcache", "-c", dest="mode", action="store_const", const="combined",
                            help="alias for --mode=combined; generate both I$ and D$ and interpret HEX_FILE input file as I$ concatenated with D$ (first half of file is I$, second half is D$)")
    output_gen_group.add_argument("--no-hex-file-addresses", action="store_true",
                            help="Suppress @-addresses in hex output.")

    subdirs_group = parser.add_argument_group("Output Directory Names")
    subdirs_group.add_argument("--icache-subdir", default=defaults['CFG_CACHE_HEX_FILES_SUBDIRS_ICACHE'])
    subdirs_group.add_argument("--dcache-subdir", default=defaults['CFG_CACHE_HEX_FILES_SUBDIRS_DCACHE'])
    subdirs_group.add_argument("--cachelines-subdir", default=defaults['CFG_CACHE_HEX_FILES_SUBDIRS_CACHELINES'])
    subdirs_group.add_argument("--tagram-subdir", default=defaults['CFG_CACHE_HEX_FILES_SUBDIRS_TAGRAM'])

    readme_group = parser.add_argument_group("Cache README Generation")
    readme_group.add_argument("--cache-readme", default=CACHE_README_FILENAME, help="filename or path of the README file to generate or append to")

    misc_group = parser.add_argument_group("Miscellaneous")
    misc_group.add_argument("--show-config", "-C", action="store_true", help="Show cache configuration and address layout.")
    misc_group.add_argument("-v", "--verbosity", "--verbose", action="count", default=0, help="Verbose output (repeat up to 2).")

    return parser

def parse_args(argv=None):
    argv = sys.argv[1:] if argv is None else argv

    # phase 1: peek at -e without consuming --help
    env_path, parent = _preparse_env(argv)
    if env_path:
        env_file_cfg = load_env_into_cfg(env_path)
    else:
        env_file_cfg = None

    # phase 2: build full parser with env-backed defaults and parse everything (incl --help)
    defaults = get_defaults(env_file_cfg)
    parser = build_main_parser(defaults, parent)
    args = parser.parse_args(argv)

    # validations that argparse can't express directly
    if not os.path.exists(args.output_dir):
        parser.error(f"output directory {args.output_dir} does not exist")
    if not os.path.exists(args.HEX_FILE):
        parser.error(f"input hex file {args.HEX_FILE} does not exist")
    if args.num_sets not in (4, 8, 16, 32, 64, 128):
        parser.error(f"num-sets must be a power of 2 between 4 and 128, got {args.num_sets}")
    if not (10 <= args.address_width <= 32):
        parser.error(f"address-width must be between 10 and 32, got {args.address_width}")

    if args.verbosity >= 1:
        if args.verbosity >= 2:
            if env_file_cfg is not None:
                print("ENVIRONMENT VARIABLES + .env FILE:")
                for k, v in env_file_cfg.items():
                    print(f"  {k}: {v}")
                print()
            else:
                print("ENVIRONMENT VARIABLES (no .env file supplied):")
                for k, v in defaults.items():
                    print(f"  {k}: {v}")
                print()
        print("FINAL ARGUMENTS (after env and CLI overrides):")
        for k, v in args.__dict__.items():
            print(f"  {k}: {v}")
        print()

    return args



import click

###############################################################################
#
# Common flags: overlay-related flags
#
###############################################################################

def overlay_opts():
    overlay_dir_opt = click.option(
        "--overlay-dir",
        "overlay_dir",
        metavar="<overlay-dir>",
        default=".",
        show_default=True,
        help="Lowest directory to look in for an overlay TOML file, after which we walk up the hierarchy.",
    )
    overlay_toml_base_name_opt = click.option(
        "--overlay-toml-base-name",
        "overlay_toml_base_name",
        metavar="<overlay-toml-base-name>",
        default="overlay.toml",
        show_default=True,
        help="Base name of overlay TOML files, e.g., 'overlay.toml' or 'dev1.overlay.toml' when using prefixes.",
    )
    overlay_prefix_opt = click.option(
        "--overlay-prefix",
        "overlay_prefix",
        metavar="<overlay-prefix>",
        default="",
        show_default=True,
        help="Prefix to select between prefixed overlay TOML files (e.g., 'dev1' selects 'dev1.overlay.toml' and 'overlay.toml')",
    )
    no_combine_overlays_opt = click.option(
        "--combine-overlays/--no-combine-overlays",
        "combine_overlays",
        is_flag=True,
        default=True,
        help="If both <prefix>.overlay.toml and overlay.toml exist, we normally apply both with the prefixed file taking precedence. This option disables that behavior and uses the prefixed file only when both are present.",
    )
    no_ascend_dir_hierarchy_opt = click.option(
        "--ascend-dir-hierarchy/--no-ascend-dir-hierarchy",
        "ascend_dir_hierarchy",
        is_flag=True,
        default=True,
        help=(
            "Do not ascend directories when searching for overlay toml files; "
            "only consider the overlay directory."
        ),
    )

    opts = [overlay_dir_opt, overlay_toml_base_name_opt, overlay_prefix_opt, no_combine_overlays_opt, no_ascend_dir_hierarchy_opt]
    
    # Apply in reverse so the first listed ends up nearest the function
    def _wrap(f):
        for opt in reversed(opts):
            f = opt(f)
        return f
    
    return _wrap



from curvtools.cli.curvcfg.cli_helpers.opts.fs_path_opt import make_fs_path_param_type_class, FsPathType
import click
from curvtools.cli.curvcfg.lib.globals.constants import GET_DEFAULT_SCHEMA_TOML_PATH

def must_exist_callback(ctx: click.Context, param: click.Parameter, value: FsPathType) -> FsPathType:
    if not value.exists():
        raise click.ClickException(f"Schema TOML file '{value}' does not exist")
    return value

def schema_file_opt(required: bool = False) -> click.Option:
    type_obj = make_fs_path_param_type_class(
            must_be_dir=False, 
            must_be_file=True, 
            default_value=GET_DEFAULT_SCHEMA_TOML_PATH())
    return click.option(
        "--schema-file",
        "schema_file",
        metavar="<schema-file>",
        default=GET_DEFAULT_SCHEMA_TOML_PATH(),
        show_default=True,
        required=required,
        help="Path to schema TOML file",
        type=type_obj,
        shell_complete=type_obj.shell_complete,
        callback=must_exist_callback,
    )



import os
from pathlib import Path
from typing import Dict, Union, Optional
from rich import get_console
from curvpyutils.file_utils.repo_utils import get_git_repo_root
from curvtools.cli.curvcfg.lib.globals.curvpaths_temporary import get_curv_root_dir_from_repo_root
from dotenv import dotenv_values
from curvtools.cli.curvcfg.lib.globals.types import CurvCliArgs
from click import Context
from click.core import ParameterSource

curvpaths: Optional[Dict[str, Path]] = None
_curvroot_dir_source: Optional[ParameterSource] = None

class CurvPaths(dict[str, Path]):
    def __init__(self, curv_root_dir: str):
        super().__init__()
        self._set_paths(curv_root_dir)
        
    def _set_paths(self, curv_root_dir: str):
        env_path = os.path.join(curv_root_dir, "scripts", "make", "paths_raw.env")
        if not os.path.isfile(env_path):
            get_console().print(f"[red]error:[/red] paths environment file not found: {env_path}")
            raise SystemExit(1)
        if "CURV_ROOT_DIR" not in os.environ:
            os.environ["CURV_ROOT_DIR"] = curv_root_dir
        env_dict = dotenv_values(dotenv_path=env_path, interpolate=True, encoding="utf-8", verbose=True)
        self["CURV_ROOT_DIR"] = Path(curv_root_dir)
        self["CURV_REPO_DIR"] = Path(env_dict.get("CURV_REPO_DIR", f"{curv_root_dir}/../..")).resolve()
        self["CURV_CONFIG_DIR"] = Path(env_dict.get("CURV_CONFIG_DIR", f"{curv_root_dir}/config")).resolve()
        self["CURV_RISCV_DIR"] = Path(env_dict.get("CURV_RISCV_DIR", f"{curv_root_dir}/riscv")).resolve()
        self["CURV_COMMON_DIR"] = Path(env_dict.get("CURV_COMMON_DIR", f"{curv_root_dir}/../common")).resolve()
        self["CURV_FIRMWARE_DIR"] = Path(env_dict.get("CURV_FIRMWARE_DIR", f"{curv_root_dir}/riscv/firmware")).resolve()

    def __str__(self):
        s = ""
        for k,v in self.items():
            s += f"{k}: {v}\n"
        s = s[:-1]
        return s

    def get_config_dir(self, add_trailing_slash: bool = False) -> str:
        s = str(self["CURV_CONFIG_DIR"])
        if add_trailing_slash and not s.endswith(os.path.sep):
            s += os.path.sep
        return s

    def get_curv_root_dir(self, add_trailing_slash: bool = False) -> str:
        s = str(self["CURV_ROOT_DIR"])
        if add_trailing_slash and not s.endswith(os.path.sep):
            s += os.path.sep
        return s

    def get_repo_dir(self, add_trailing_slash: bool = False) -> str:
        s = str(self["CURV_REPO_DIR"])
        if add_trailing_slash and not s.endswith(os.path.sep):
            s += os.path.sep
        return s

    # def get_schema_toml_path(self) -> str:
    #     return str(Path(self["CURV_CONFIG_DIR"]) / "schema" / "schema.toml")

    def _try_make_relative_to_dir(self,path: str, dir: str) -> str:
        """
        Try to make an absolute path into a path relative to a directory.
        If the path is not relative to the directory, return the absolute path.
        """
        # Convert both paths to absolute paths
        abs_path = Path(path).absolute()
        abs_dir = Path(dir).absolute()
        if not abs_path.is_relative_to(abs_dir):
            return str(abs_path)
        else:
            return str(abs_path.relative_to(abs_dir))

    def mk_rel_to_curv_root(self, abs_path: str) -> str:
        """
        Try to make an absolute path into a path relative to the curv root directory.
        If the path is not relative to the curv root directory, return the absolute path.
        """
        return self._try_make_relative_to_dir(abs_path, self.get_curv_root_dir())

    def mk_rel_to_cwd(self, abs_path: str) -> str:
        """
        Try to make an absolute path into a path relative to a directory.
        If the path is not relative to the directory, return the absolute path.
        """
        return self._try_make_relative_to_dir(abs_path, os.getcwd())

    def mk_rel_to_curv_config_dir(self, abs_path: str) -> str:
        """
        Try to make an absolute path into a path relative to the curv config directory.
        If the path is not relative to the curv config directory, return the absolute path.
        """
        return self._try_make_relative_to_dir(abs_path, self.get_config_dir())

def get_curv_paths(args: CurvCliArgs | Context | None = None) -> CurvPaths:
    """
    Get the paths commonly used in this build system, and track where CURV_ROOT_DIR was obtained from.
    """
    global curvpaths, _curvpaths_source

    # initialize curvpaths if it's not already initialized
    # (if we get called a second time with a non-None args, we re-initialize)
    if curvpaths is None or args is not None:

        # reset if it was previously set since now we're updating
        _curvroot_dir_source = None

        # extract args if we were passed a Context
        if isinstance(args, Context):
            temp_args = args.obj
        else:
            temp_args = args

        # if args was supplied, try it, but always fall back to the environment variable if set
        if temp_args is None or "curv_root_dir" not in temp_args:
            curv_root_dir = os.environ.get("CURV_ROOT_DIR")
            if curv_root_dir is not None and os.path.isdir(curv_root_dir):
                _curvroot_dir_source = ParameterSource.ENVIRONMENT
        else:
            curv_root_dir = temp_args.get("curv_root_dir", None)
            if curv_root_dir is not None and os.path.isdir(curv_root_dir):
                _curvroot_dir_source = ParameterSource.COMMANDLINE
            else:
                curv_root_dir = os.environ.get("CURV_ROOT_DIR")
                if curv_root_dir is not None and os.path.isdir(curv_root_dir):
                    _curvroot_dir_source = ParameterSource.ENVIRONMENT

        # fall back to git rev-parse to find repo root if unable to get CURV_ROOT_DIR from args or environment variable
        if not curv_root_dir:
            repo_root_dir = get_git_repo_root()
            curv_root_dir = get_curv_root_dir_from_repo_root(repo_root_dir)
            curv_root_dir = os.path.expanduser(str(curv_root_dir))

        # fail if we have not gotten a valid dir yet for CURV_ROOT_DIR
        if not os.path.isdir(curv_root_dir):
            get_console().print(f"[red]error:[/red] --curv-root-dir not found: {curv_root_dir}")
            raise SystemExit(1)
        else:
            # Default in this context means it was found through git-rev-parse
            _curvroot_dir_source = ParameterSource.DEFAULT

        curvpaths = CurvPaths(curv_root_dir)

    return curvpaths


import os
import filecmp
import json
import re
from enum import Flag
from typing import Dict, Optional, Any
from . import CfgValues, CfgValue

class ConfigFileTypes(Flag):
    NONE = 0
    ENV = 1
    SVPKG = 2
    SVH = 4
    MAKEFILE = 8
    JSON = 16
    ALL = ENV | SVPKG | SVH | MAKEFILE | JSON

ConfigFileTypesForWriting = ConfigFileTypes.MAKEFILE | ConfigFileTypes.ENV | ConfigFileTypes.SVPKG | ConfigFileTypes.SVH

DEFAULT_OUTFILE_NAMES = {
    "makefile": "curv.mk",
    "env": ".curv.env",
    "svh": "curvcfg.svh",
    "svpkg": "curvcfgpkg.sv",
    "json": "curvcfg.json",
}

class FormatUtils:
    @staticmethod
    def extract_sv_width(sv_type:Optional[str]) -> Optional[int]:
        # Expect forms like "logic [31:0]" or "wire [7:0]"
        if not sv_type:
            return None
        m = re.search(r"\[(\d+)\s*:\s*0\]", sv_type)
        return int(m.group(1)) + 1 if m else None

    @staticmethod
    def sv_numeric_literal(value:int, sv_type: Optional[str] = None, makefile_type: Optional[str] = None) -> str:
        w = FormatUtils.extract_sv_width(sv_type)
        if w is not None:
            hexdigs = (w + 3) // 4
            return "{w}'h{v:0{d}X}".format(w=w, v=value, d=hexdigs)
        # Fallback 1:  use makefile_type as hint for hex width
        if makefile_type is not None:
            if makefile_type == "hex":
                return "32'h{0:08X}".format(value)
            elif makefile_type == "hex32":
                return "32'h{0:08X}".format(value)
            elif makefile_type == "hex16":
                return "16'h{0:04X}".format(value)
            elif makefile_type == "hex8":
                return "8'h{0:02X}".format(value)
            elif makefile_type == "decimal":
                return str(value)
            elif makefile_type == "string":
                return str(value)
        # Fallback 2: infer from value's magnitude
        if value < 0:
            return "32'h{0:08X}".format(value)
        elif value > 1<<8:
            return "8'h{0:02X}".format(value)
        elif value > 1<<16:
            return "16'h{0:04X}".format(value)
        # Fallback 3: default to 32-bit hex literal
        return "32'h{0:08X}".format(value)

    @staticmethod
    def format_make_value(v:Any, makefile_type: Optional[str] = None) -> str:
        if isinstance(v, int):
            if makefile_type == "hex":
                return "0x{0:X}".format(v)
            elif makefile_type == "hex32":
                return "0x{0:08X}".format(v)
            elif makefile_type == "hex16":
                return "0x{0:04X}".format(v)
            elif makefile_type == "hex8":
                return "0x{0:02X}".format(v)
            else: # decimal or string type
                return str(v)
        elif isinstance(v, str):
            if makefile_type == "hex":
                return "0x{0:X}".format(int(v))
            elif makefile_type == "hex32":
                return "0x{0:08X}".format(int(v))
            elif makefile_type == "hex16":
                return "0x{0:04X}".format(int(v))
            elif makefile_type == "hex8":
                return "0x{0:02X}".format(int(v))
            elif makefile_type == "decimal":
                return str(int(v))
            else:
                return v
        return str(v)
        
class FileEmitter:
    def __init__(self, config_values:CfgValues, outdir_path:str, emit_files: ConfigFileTypes, verbosity: int = 0):
        self.config_values = config_values
        self.outdir_path = outdir_path
        self.emit_files = emit_files
        self.verbosity = verbosity
        self.output_paths = {"makefile": os.path.join(self.outdir_path, DEFAULT_OUTFILE_NAMES["makefile"]),
                            "env": os.path.join(self.outdir_path, DEFAULT_OUTFILE_NAMES["env"]),
                            "svpkg": os.path.join(self.outdir_path, DEFAULT_OUTFILE_NAMES["svpkg"]),
                            "svh": os.path.join(self.outdir_path, DEFAULT_OUTFILE_NAMES["svh"]),
                            "json": os.path.join(self.outdir_path, DEFAULT_OUTFILE_NAMES["json"]),
                            }
        self.emitter_functions = {  "makefile": self._emit_makefile,
                                    "env": self._emit_env_file,
                                    "svpkg": self._emit_sv_pkg,
                                    "svh": self._emit_svh_defines,
                                    "json": self._emit_json,
                                    }

    def _ensure_outdir(self, p:str) -> None:
        os.makedirs(p, exist_ok=True)

    def emit(self, write_only_if_changed: bool = True) -> tuple[list[str], list[str]]:        
        # make list of files we want to emit
        files_to_emit = []
        if self.emit_files & ConfigFileTypes.MAKEFILE:
            files_to_emit.append("makefile")
        if self.emit_files & ConfigFileTypes.ENV:
            files_to_emit.append("env")
        if self.emit_files & ConfigFileTypes.SVPKG:
            files_to_emit.append("svpkg")
        if self.emit_files & ConfigFileTypes.SVH:
            files_to_emit.append("svh")
        if self.emit_files & ConfigFileTypes.JSON:
            files_to_emit.append("json")

        files_emitted: list[str] = []    
        files_unchanged: list[str] = []        
        for key in files_to_emit:
            was_overwritten = self.emitter_functions[key](write_only_if_changed)
            if was_overwritten:
                if self.verbosity >= 2:
                    print(f"⭐️ Updated or created: {self.output_paths[key]}")
                files_emitted.append(self.output_paths[key])
            else:
                if self.verbosity >= 2:
                    print(f"(no changes to {self.output_paths[key]})")
                files_unchanged.append(self.output_paths[key])
        return files_emitted, files_unchanged

    def _is_file_changed(self, temp_output_path: str, output_path: str) -> bool:
        if os.path.exists(output_path):
            if filecmp.cmp(temp_output_path, output_path):
                return False
        return True

    def _emit_makefile(self, write_only_if_changed: bool = True) -> bool:
        """
        Emits a Makefile with the config values into <output_path>.

        - If the Makefile does not exist, it is created.
        - If the Makefile already exists, it is only overwritten if the contents are different.

        Returns:
        True if the Makefile was overwritten, False if it was not.
        """
        output_path = self.output_paths["makefile"]
        self._ensure_outdir(os.path.dirname(output_path))
        temp_output_path = output_path + ".tmp"
        with open(temp_output_path, "w") as f:
            f.write("# Autogenerated by curvcfg.py. Do not edit.\n")
            for k in sorted(self.config_values.keys()):
                v_obj: CfgValue = self.config_values[k]
                locs = v_obj.meta.locations
                if "all" in locs or "makefiles" in locs:
                    raw = v_obj.get_raw_value()
                    f.write("{k} := {v}\n".format(k=k, v=FormatUtils.format_make_value(raw, v_obj.meta.makefile_type)))
        
        # Check if the file changed
        file_changed = self._is_file_changed(temp_output_path, output_path)
        
        # If the file changed or we are not writing only if changed, overwrite the output file
        if (not write_only_if_changed) or file_changed:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            os.rename(temp_output_path, output_path)
            return True
        else:
            os.remove(temp_output_path)
            return False

    def _emit_env_file(self, write_only_if_changed: bool = True) -> bool:
        """
        Emits a shell environment file with the config values into <output_path>.

        - If the shell environment file does not exist, it is created.
        - If the shell environment file already exists, it is only overwritten if the contents are different.

        Returns:
        True if the shell environment file was overwritten, False if it was not.
        """
        output_path = self.output_paths["env"]
        self._ensure_outdir(os.path.dirname(output_path))
        temp_output_path = output_path + ".tmp"
        with open(temp_output_path, "w") as f:
            f.write("# -----------------------------------------------------------------------------\n")
            f.write("# Autogenerated by curvcfg.py. Do not edit.\n")
            f.write("# -----------------------------------------------------------------------------\n\n")
            for k in sorted(self.config_values.keys()):
                v_obj: CfgValue = self.config_values[k]
                locs = v_obj.meta.locations
                if "all" in locs or "env" in locs:
                    raw = v_obj.get_raw_value()
                    f.write("{k}={v}\n".format(k=k, v=FormatUtils.format_make_value(raw, v_obj.meta.makefile_type)))

        # Check if the file changed
        file_changed = self._is_file_changed(temp_output_path, output_path)
        
        # If the file changed or we are not writing only if changed, overwrite the output file
        if (not write_only_if_changed) or file_changed:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            os.rename(temp_output_path, output_path)
            return True
        else:
            os.remove(temp_output_path)
            return False

    def _emit_svh_defines(self, write_only_if_changed: bool = True) -> bool:
        """
        Emits a SystemVerilog include header file with the config values into <output_path>.

        - If the SystemVerilog include header file does not exist, it is created.
        - If the SystemVerilog include header file already exists, it is only overwritten if the contents are different.

        Returns:
        True if the SystemVerilog include header file was overwritten, False if it was not.
        """
        output_path = self.output_paths["svh"]
        self._ensure_outdir(os.path.dirname(output_path))
        temp_output_path = output_path + ".tmp"
        guard = "__CURVCFG_SVH__"
        with open(temp_output_path, "w") as f:
            f.write("// -----------------------------------------------------------------------------\n")
            f.write("// Autogenerated by curvcfg.py. Do not edit.\n")
            f.write("// -----------------------------------------------------------------------------\n\n")
            f.write("`ifndef {g}\n`define {g}\n\n".format(g=guard))
            any_defs = False
            longest_key_len = max(len(k) for k in self.config_values.keys())
            for k in sorted(self.config_values.keys()):
                v_obj: CfgValue = self.config_values[k]
                locs = v_obj.meta.locations
                if "all" in locs or "defines" in locs:
                    any_defs = True
                    sv_type = v_obj.meta.sv_type
                    makefile_type = v_obj.meta.makefile_type
                    vtype   = v_obj.meta.type
                    v       = v_obj.get_raw_value()
                    if (vtype  == "int") or (vtype == "enum" and isinstance(v, int)):
                        if (sv_type == "int"):
                            lit = v
                        else:
                            lit = FormatUtils.sv_numeric_literal(v, sv_type, makefile_type)
                        f.write("`define {k} {lit}\n".format(k=k.ljust(longest_key_len+4), lit=lit))
                    elif (vtype == "string") or (vtype == "enum" and isinstance(v, str)):
                        f.write("`define {k} `\"{v}`\"\n".format(k=k.ljust(longest_key_len+4), v=v))
                    elif isinstance(v, int):
                        f.write("`define {k} 32'h{v:08X}\n".format(k=k.ljust(longest_key_len+4), v=v))
                    else:
                        # Strings can be awkward in macro context; still emit something useful.
                        f.write("// {k_comment} = \"{v}\"\n`define {k} `\"{v}`\"\n".format(k_comment=k,k=k.ljust(longest_key_len+4), v=v))
            if not any_defs:
                f.write("// (No defines selected by schema locations)\n")
            f.write("\n`endif // {g}\n".format(g=guard))

        # Check if the file changed
        file_changed = self._is_file_changed(temp_output_path, output_path)
        
        # If the file changed or we are not writing only if changed, overwrite the output file
        if (not write_only_if_changed) or file_changed:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            os.rename(temp_output_path, output_path)
            return True
        else:
            os.remove(temp_output_path)
            return False

    def _emit_sv_pkg(self, write_only_if_changed: bool = True) -> bool:
        """
        Emits a SystemVerilog package file with the config values into <output_path>.

        - If the SystemVerilog package file does not exist, it is created.
        - If the SystemVerilog package file already exists, it is only overwritten if the contents are different.

        Returns:
        True if the SystemVerilog package file was overwritten, False if it was not.
        """
        output_path = self.output_paths["svpkg"]
        self._ensure_outdir(os.path.dirname(output_path))
        temp_output_path = output_path + ".tmp"
        with open(temp_output_path, "w") as f:
            f.write("// -----------------------------------------------------------------------------\n")
            f.write("// Autogenerated by curvcfg.py. Do not edit.\n")
            f.write("// -----------------------------------------------------------------------------\n\n")
            f.write("package curvcfgpkg;\n\n")
            f.write("  // verilator lint_off UNUSEDPARAM\n")
            for k in sorted(self.config_values.keys()):
                v_obj: CfgValue = self.config_values[k]
                locs = v_obj.meta.locations
                if "all" in locs or "cfgpkg" in locs:
                    sv_type = v_obj.meta.sv_type
                    makefile_type = v_obj.meta.makefile_type
                    vtype   = v_obj.meta.type
                    v       = v_obj.get_raw_value()
                    if (vtype  == "int") or (vtype == "enum" and isinstance(v, int)):
                        if (sv_type == "int"):
                            lit = v
                        else:
                            lit = FormatUtils.sv_numeric_literal(v, sv_type, makefile_type)
                        f.write("  localparam {t} {k} = {lit};\n".format(t=sv_type, k=k, lit=lit))
                    elif (vtype == "string") or (vtype == "enum" and isinstance(v, str)):
                        f.write("  localparam string {k} = \"{v}\";\n".format(k=k, v=v))
                    else:
                        # Fallback types
                        if isinstance(v, int):
                            f.write("  localparam int {k} = {v};\n".format(k=k, v=v))
                        else:
                            f.write("  localparam string {k} = \"{v}\";\n".format(k=k, v=v))
            f.write("  // verilator lint_on UNUSEDPARAM\n\n")
            f.write("endpackage : curvcfgpkg\n")

        # Check if the file changed
        file_changed = self._is_file_changed(temp_output_path, output_path)
        
        # If the file changed or we are not writing only if changed, overwrite the output file
        if (not write_only_if_changed) or file_changed:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            os.rename(temp_output_path, output_path)
            return True
        else:
            os.remove(temp_output_path)
            return False

    def _emit_json(self, write_only_if_changed: bool = True) -> bool:
        """
        Emits a JSON file with the config values into <output_path>.

        - If the JSON file does not exist, it is created.
        - If the JSON file already exists, it is only overwritten if the contents are different.

        Returns:
        True if the JSON file was overwritten, False if it was not.
        """
        output_path = self.output_paths["json"]
        self._ensure_outdir(os.path.dirname(output_path))
        temp_output_path = output_path + ".tmp"
        flat:Dict[str, str | int | None] = {}
        for k in sorted(self.config_values.keys()):
            v_obj: CfgValue = self.config_values[k]
            locs = v_obj.meta.locations
            if "all" in locs or "json" in locs:
                flat[k] = v_obj.get_raw_value()
        with open(temp_output_path, "w") as f:
            json.dump(flat, f, indent=2, sort_keys=True)

        # Check if the file changed
        file_changed = self._is_file_changed(temp_output_path, output_path)
        
        # If the file changed or we are not writing only if changed, overwrite the output file
        if (not write_only_if_changed) or file_changed:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
            os.rename(temp_output_path, output_path)
            return True
        else:
            os.remove(temp_output_path)
            return False

    # def _emit_config_mk_deps(self, write_only_if_changed: bool = True) -> bool:
    #     """
    #     Emits a Makefile dependencies file that can be included to indicate that each of our
    #     emitted config files depends on the list of toml files that were used to generate them.

    #     - If the Makefile dependencies file does not exist, it is created.
    #     - If the Makefile dependencies file already exists, it is only overwritten if the contents are different.
    #     Unless write_only_if_changed is False, in which case the file is always overwritten.

    #     Returns:
    #     True if the Makefile dependencies file was overwritten, False if it was not.
    #     """
    #     output_path = self.output_paths["config_mk_deps"]
    #     self._ensure_outdir(os.path.dirname(output_path))
    #     temp_output_path = output_path + ".tmp"
    #     deps:list[str] = []
    #     if self.emit_files & ConfigFileTypes.MAKEFILE:
    #         deps.append(self.output_paths["makefile"])
    #     if self.emit_files & ConfigFileTypes.ENV:
    #         deps.append(self.output_paths["env"])
    #     if self.emit_files & ConfigFileTypes.SVPKG:
    #         deps.append(self.output_paths["svpkg"])
    #     if self.emit_files & ConfigFileTypes.SVH:
    #         deps.append(self.output_paths["svh"])
    #     if self.emit_files & ConfigFileTypes.JSON:
    #         deps.append(self.output_paths["json"])
    #     if self.emit_files & ConfigFileTypes.CONFIG_MK_DEPS:
    #         deps.append(self.output_paths["config_mk_deps"])

    #     target_list_str = ' '.join(deps)
    #     toml_list_str = ' '.join(self.tomls_abs_paths)

    #     with open(temp_output_path, "w") as f:
    #         f.write(f"{target_list_str}: {toml_list_str}\n")

    #     # Check if the file changed
    #     file_changed = self._is_file_changed(temp_output_path, output_path)
        
    #     # If the file changed or we are not writing only if changed, overwrite the output file
    #     if (not write_only_if_changed) or file_changed:
    #         os.rename(temp_output_path, output_path)
    #         return True
    #     else:
    #         os.remove(temp_output_path)
    #         return False



import click

###############################################################################
#
# common flags: --verbose/-v
#
###############################################################################
def verbosity_opts(include_verbose: bool):
    def cap3(ctx: click.Context, _param: click.Parameter, value: int) -> int: 
        ctx.ensure_object(dict)
        if "verbosity" not in ctx.obj:
            ctx.obj["verbosity"] = 0
        ctx.obj["verbosity"] = max(ctx.obj["verbosity"], min(value, 3))
        return ctx.obj["verbosity"]
    
    verbose_option = click.option(
        "--verbose", "-v",
        "verbose",
        count=True,
        help="Enables verbose mode. Repeat for more verbosity (up to -vvv).",
        callback=cap3,
    )
    opts = []
    opts.append(verbose_option) if include_verbose else None
    opts = [opt for opt in opts if opt is not None]

    # Apply in reverse so the first listed ends up nearest the function
    def _wrap(f):
        for opt in reversed(opts):
            f = opt(f)
        return f
    return _wrap



#!/usr/bin/env python3
"""
Unit tests for the memory map validator
"""
import pytest
from curvtools.cli.memmap2.validator import MemoryMapValidator, ValidationSeverity

pytestmark = [pytest.mark.unit]

def test_valid_memory_map():
    """Test validation on a valid memory map"""
    valid_memory_map = {
        "slaves": {
            "sdram": {
                "ranges": [
                    {"name": "RAM", "start": 0x00000000, "end": 0x01FFFFFF, "access": "rw", "cacheable": True}
                ]
            }
        }
    }

    validator = MemoryMapValidator()
    errors = validator.validate(valid_memory_map, xlen=32)

    assert len(errors) == 0, f"Expected no errors, got: {[e.message for e in errors]}"

def test_overlapping_ranges():
    """Test detection of overlapping ranges"""
    invalid_memory_map = {
        "slaves": {
            "test": {
                "ranges": [
                    {"name": "Range1", "start": 0x00000000, "end": 0x0000000F, "access": "rw", "cacheable": False},
                    {"name": "Range2", "start": 0x00000008, "end": 0x00000017, "access": "rw", "cacheable": False}
                ]
            }
        }
    }

    validator = MemoryMapValidator()
    errors = validator.validate(invalid_memory_map, xlen=32)

    overlap_errors = [e for e in errors if e.rule_id == "no_overlaps"]
    assert len(overlap_errors) > 0, "Should detect overlapping ranges"

def test_overlapping_registers():
    """Test detection of overlapping registers with XLEN-dependent sizing"""
    # Create a slave with a range containing two adjacent 4-byte spaced registers
    memory_map = {
        "slaves": {
            "test": {
                "ranges": [
                    {"name": "Test Range", "start": 0x00000000, "end": 0x0000000F, "cacheable": False}
                ],
                "registers": {
                    "reg1": {"name": "Register 1", "addr": 0x00000000, "access": "rw"},
                    "reg2": {"name": "Register 2", "addr": 0x00000004, "access": "rw"}
                }
            }
        }
    }

    validator = MemoryMapValidator()

    # With XLEN=32, registers should NOT overlap (4-byte registers, 4-byte spacing)
    errors_32 = validator.validate(memory_map, xlen=32)
    register_overlap_errors_32 = [e for e in errors_32 if e.rule_id == "no_register_overlaps"]
    assert len(register_overlap_errors_32) == 0, f"XLEN=32 should not have register overlaps, got: {[e.message for e in register_overlap_errors_32]}"

    # With XLEN=64, registers SHOULD overlap (8-byte registers, 4-byte spacing)
    errors_64 = validator.validate(memory_map, xlen=64)
    register_overlap_errors_64 = [e for e in errors_64 if e.rule_id == "no_register_overlaps"]
    assert len(register_overlap_errors_64) > 0, "XLEN=64 should detect register overlaps"

def test_invalid_access_value():
    """Test detection of invalid access values"""
    invalid_memory_map = {
        "slaves": {
            "test": {
                "ranges": [
                    {"name": "Test", "start": 0x00000000, "end": 0x00000003, "access": "invalid", "cacheable": False}
                ]
            }
        }
    }

    validator = MemoryMapValidator()
    errors = validator.validate(invalid_memory_map, xlen=32)

    access_errors = [e for e in errors if e.rule_id == "valid_access_values"]
    assert len(access_errors) > 0, "Should detect invalid access values"

def test_word_alignment():
    """Test word alignment validation"""
    invalid_memory_map = {
        "slaves": {
            "test": {
                "ranges": [
                    {"name": "Test", "start": 0x00000001, "end": 0x00000004, "access": "rw", "cacheable": False}
                ]
            }
        }
    }

    validator = MemoryMapValidator()
    errors = validator.validate(invalid_memory_map, xlen=32)

    alignment_errors = [e for e in errors if e.rule_id == "word_aligned"]
    assert len(alignment_errors) > 0, "Should detect unaligned addresses"

def test_highest_cacheable_address():
    """Test calculation of highest cacheable address"""
    memory_map = {
        "slaves": {
            "sdram": {
                "ranges": [
                    {"name": "RAM", "start": 0x00000000, "end": 0x01FFFFFF, "cacheable": True}
                ]
            },
            "flash": {
                "ranges": [
                    {"name": "ROM", "start": 0x02000000, "end": 0x02FFFFFF, "access": "ro"}
                ]
            }
        }
    }

    validator = MemoryMapValidator()
    highest = validator.get_highest_cacheable_address(memory_map, xlen=32)

    assert highest == 0x01FFFFFF, f"Expected highest cacheable address 0x01FFFFFF, got 0x{highest:08x}"



#!/usr/bin/env python3
"""
Default Jinja2 template for SystemVerilog package generation
"""

DEFAULT_JINJA2_TEMPLATE = """//
//   This file is automatically generated by {{ script_name }}. See MEMORY_MAP.md for more information.
//
//
{{ memmap_comment_str }}
//
//
//   >>>> Memory Mapped Registers and Regions <<<
//
{{ registers_and_sub_ranges_comment_str }}
//

package memmappkg;

// verilator lint_off UNUSEDPARAM

localparam {{ "%-35s"|format("WB_SSEL_WIDTH") }} = {{ enum_width }};
{{ "%-64s"|format("localparam " ~ "%-35s"|format("WB_SLV_CNT") ~ " = WB_SSEL_WIDTH;") }}// equal since ssel is one-hot

// Highest cacheable address (for cache tag optimization)
localparam {{ "%-35s"|format("HIGHEST_CACHEABLE_ADDRESS") }} = {{ highest_addr_cacheable_str }};

// Slave select constants (one-hot encoded)
{% for slave in slaves -%}
{{ "%-64s"|format("localparam " ~ "%-35s"|format("WB_SLAVE_" ~ slave.slave_name.upper()) ~ " = " ~ slave.format_one_hot(slave.one_hot, enum_width) ~ ";") }}// bit position {{ slave.bitpos }}
{% endfor -%}
{{ "%-64s"|format("localparam " ~ "%-35s"|format("WB_SLAVE_NONE") ~ " = " ~ wb_slave_none_value_str ~ ";") }}// none selected

// ------------------------------------------------------------------------------------
// Functions for address translation
// ------------------------------------------------------------------------------------

{% if use_inside_range_condition %}
{% include 'get_slave_sel_inside.sv.j2' %}
{% else %}
{% include 'get_slave_sel_or_chain.sv.j2' %}
{% endif %}

// ------------------------------------------------------------------------------------
// Functions for slaves
// ------------------------------------------------------------------------------------

`ifdef SIMULATION

// Slave can use these to determine whether it is selected by the given address
{% for slave in slaves -%}
localparam {{ "%-35s"|format(slave.slave_name.lower()+"_bitpos") }} = {{ slave.bitpos }};
{% endfor %}

{% for slave in slaves %}
function logic is_selected_slave_{{ slave.slave_name.lower() }} (input logic [31:0] addr, input logic [{{ enum_width-1 }}:0] slave_sel_i);
    is_selected_slave_{{ "%-24s"|format(slave.slave_name.lower()) }} = slave_sel_i[{{ slave.slave_name.lower()+"_bitpos" }}];
endfunction
{% endfor %}

`ifdef SLANG
$static_assert(is_selected_slave_sdram(32'h0000_0000, WB_SLAVE_SDRAM), "WB_SLAVE_SDRAM should cause selected slave to be sdram");
$static_assert(!is_selected_slave_sdram(32'h0000_0000, WB_SLAVE_FLASH), "WB_SLAVE_FLASH should not cause selected slave to be sdram");
$static_assert(!is_selected_slave_sdram(32'h0000_0000, WB_SLAVE_BRAM), "WB_SLAVE_BRAM should not cause selected slave to be sdram");
$static_assert(!is_selected_slave_sdram(32'h0000_0000, WB_SLAVE_PERIPHERALS), "WB_SLAVE_PERIPHERALS should not cause selected slave to be sdram");
$static_assert(!is_selected_slave_sdram(32'h0000_0000, WB_SLAVE_NONE), "WB_SLAVE_NONE should not cause selected slave to be sdram");

$static_assert(!is_selected_slave_flash(32'h0000_0000, WB_SLAVE_SDRAM), "WB_SLAVE_SDRAM should not cause selected slave to be flash");
$static_assert(is_selected_slave_flash(32'h0000_0000, WB_SLAVE_FLASH), "WB_SLAVE_FLASH should cause selected slave to be flash");
$static_assert(!is_selected_slave_flash(32'h0000_0000, WB_SLAVE_BRAM), "WB_SLAVE_BRAM should not cause selected slave to be flash");
$static_assert(!is_selected_slave_flash(32'h0000_0000, WB_SLAVE_PERIPHERALS), "WB_SLAVE_PERIPHERALS should not cause selected slave to be flash");
$static_assert(!is_selected_slave_flash(32'h0000_0000, WB_SLAVE_NONE), "WB_SLAVE_NONE should not cause selected slave to be flash");

$static_assert(!is_selected_slave_bram(32'h0000_0000, WB_SLAVE_SDRAM), "WB_SLAVE_SDRAM should not cause selected slave to be bram");
$static_assert(!is_selected_slave_bram(32'h0000_0000, WB_SLAVE_FLASH), "WB_SLAVE_FLASH should not cause selected slave to be bram");
$static_assert(is_selected_slave_bram(32'h0000_0000, WB_SLAVE_BRAM), "WB_SLAVE_BRAM should not cause selected slave to be bram");
$static_assert(!is_selected_slave_bram(32'h0000_0000, WB_SLAVE_PERIPHERALS), "WB_SLAVE_PERIPHERALS should not cause selected slave to be bram");
$static_assert(!is_selected_slave_bram(32'h0000_0000, WB_SLAVE_NONE), "WB_SLAVE_NONE should not cause selected slave to be bram");

$static_assert(!is_selected_slave_peripherals(32'h0000_0000, WB_SLAVE_SDRAM), "WB_SLAVE_SDRAM should not cause selected slave to be peripherals");
$static_assert(!is_selected_slave_peripherals(32'h0000_0000, WB_SLAVE_FLASH), "WB_SLAVE_FLASH should not cause selected slave to be peripherals");
$static_assert(!is_selected_slave_peripherals(32'h0000_0000, WB_SLAVE_BRAM), "WB_SLAVE_BRAM should not cause selected slave to be peripherals");
$static_assert(is_selected_slave_peripherals(32'h0000_0000, WB_SLAVE_PERIPHERALS), "WB_SLAVE_PERIPHERALS should cause selected slave to be peripherals");
$static_assert(!is_selected_slave_peripherals(32'h0000_0000, WB_SLAVE_NONE), "WB_SLAVE_NONE should not cause selected slave to be peripherals");
`endif

`endif

// convert one-hot to index using priority encoder
function automatic int get_slave_idx (input logic [{{ enum_width-1 }}:0] slave_sel_i);
    get_slave_idx = 0; // default to none selected
    for (int i = 0; i < WB_SLV_CNT; i++) begin
        if (slave_sel_i[i]) begin 
            get_slave_idx = i;
        end
    end
endfunction

`ifdef SLANG
$static_assert(get_slave_idx(WB_SLAVE_SDRAM) == sdram_bitpos, "WB_SLAVE_SDRAM should have index 0");
$static_assert(get_slave_idx(WB_SLAVE_FLASH) == flash_bitpos, "WB_SLAVE_FLASH should have index 1");
$static_assert(get_slave_idx(WB_SLAVE_BRAM) == bram_bitpos, "WB_SLAVE_BRAM should have index 2");
$static_assert(get_slave_idx(WB_SLAVE_PERIPHERALS) == peripherals_bitpos, "WB_SLAVE_PERIPHERALS should have index 3");
$static_assert(get_slave_idx(WB_SLAVE_NONE) == 0, "WB_SLAVE_NONE should have index 0");
`endif

// ------------------------------------------------------------------------------------
// Functions for access legality
// ------------------------------------------------------------------------------------

typedef enum logic [0:0] {
    READ_LEGAL  = 1'b0,
    WRITE_LEGAL = 1'b1
} access_mode_t;
{% for slave in slaves -%}
function logic is_legal_access_{{ slave.slave_name.lower() }}(input logic[31:0] addr, access_mode_t access);
    case (access)
        READ_LEGAL:  is_legal_access_{{ slave.slave_name.lower() }} = {% for expr in slave.read_legal_exprs -%}
                {{ expr }}
            {%- endfor -%};
        WRITE_LEGAL: is_legal_access_{{ slave.slave_name.lower() }} = {% for expr in slave.write_legal_exprs -%}
                {{ expr }}
            {%- endfor -%};
        default:     is_legal_access_{{ slave.slave_name.lower() }} = 1'b0;
    endcase
endfunction

{% endfor -%}

// ------------------------------------------------------------------------------------
// Register address constants
// ------------------------------------------------------------------------------------

{% for slave in slaves -%}
{% if slave.registers %}

//
// {{ slave.slave_name }} register addresses
//
{% for register in slave.registers -%}
localparam REG_{{ register.get_name_snake_case(False).upper().ljust(30) }} = {{ register.get_addr_hex_str(True, 4, "verilog") }};
{% endfor -%}
{% endif -%}
{% endfor %}

// ------------------------------------------------------------------------------------
// Register test functions
// ------------------------------------------------------------------------------------

{% for slave in slaves %}
{% if slave.registers %}
{% if not loop.first %}

{% endif %}
//
// {{ slave.slave_name }} register test functions
//

{% for register in slave.registers %}
{% set reg_snake = register.get_name_snake_case(False) %}
{% set fn_name = reg_snake if reg_snake.endswith('_register') else reg_snake + '_register' %}
// is addr the {{ register.name }}?
function logic is_addr_{{ fn_name }} (input logic [31:0] addr);
    is_addr_{{ fn_name }} = (addr==REG_{{ register.get_name_snake_case(False).upper() }});
endfunction
{% endfor %}

{% endif %}
{% endfor %}

// ------------------------------------------------------------------------------------
// Range and buffer test functions
// ------------------------------------------------------------------------------------

{% if use_inside_range_condition %}
{% include 'range_buffer_tests_inside.sv.j2' %}
{% else %}
{% include 'range_buffer_tests_gte_lte.sv.j2' %}
{% endif %}

// verilator lint_on UNUSEDPARAM

endpackage

"""

DEFAULT_GET_SLAVE_SEL_FUNC_OR_CHAIN = """
// Compute slave_sel_o for the given address
function logic [{{ enum_width-1 }}:0] get_slave_sel_o (input logic [31:0] addr);
    {% for s in slave_addr_ranges %}
    if (
        {%- for r in s.ranges %}
        {%- if r.start == 0 %}
        ($unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }}){{ ' ||' if not loop.last else '' }}
        {%- else %}
        ($unsigned(addr) >= 32'h{{ "%08x"|format(r.start) }} && $unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }}){{ ' ||' if not loop.last else '' }}
        {%- endif %}
        {%- endfor %}
    ) return WB_SLAVE_{{ s.slave_name.upper() }};
    {% endfor %}
    return WB_SLAVE_NONE; // No slave selected
endfunction
"""

DEFAULT_GET_SLAVE_SEL_FUNC_INSIDE = """
// Compute slave_sel_o for the given address
function logic [{{ enum_width-1 }}:0] get_slave_sel_o (input logic [31:0] addr);
    unique case ($unsigned(addr)) inside
    {% for s in slave_addr_ranges %}
    {% if s.ranges|length > 0 %}
        {% if s.ranges|length == 1 %}
        [32'h{{ "%08x"|format(s.ranges[0].start) }} : 32'h{{ "%08x"|format(s.ranges[0].end) }}]       : get_slave_sel_o = WB_SLAVE_{{ s.slave_name.upper() }};
        {% else %}
        {% for r in s.ranges %}
        [32'h{{ "%08x"|format(r.start) }} : 32'h{{ "%08x"|format(r.end) }}]{{ "," if not loop.last else "       : get_slave_sel_o = WB_SLAVE_" ~ s.slave_name.upper() ~ ";" }}
        {% endfor %}
        {% endif %}
    {% endif %}
    {% endfor %}
        default                             : get_slave_sel_o = WB_SLAVE_NONE;
    endcase
endfunction
"""

DEFAULT_RANGE_TEST_FUNCTIONS_GTE_LTE = """{% for slave in slaves_collected %}
{% if slave.ranges|length > 0 %}
{% if not loop.first %}

{% endif %}
//
// {{ slave.name }} range and buffer test functions
//

// is addr handled by {{ slave.name }} bus slave
function logic is_addr_in_{{ slave.name|lower|replace(' ', '_') }} (input logic [31:0] addr);
    is_addr_in_{{ slave.name|lower|replace(' ', '_') }} = (
        {%- for r in slave.ranges %}
        {%- if r.start == 0 -%}
        $unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }}
        {%- else -%}
        ($unsigned(addr) >= 32'h{{ "%08x"|format(r.start) }} && $unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }})
        {%- endif -%}
        {{ ' ||' if not loop.last else '' }}
        {%- endfor %}
    );
endfunction

{% for r in slave.ranges %}
// is addr in range {{ r.name }} of the {{ slave.name }} bus slave?
function logic is_addr_in_{{ r.name_snake }}_range (input logic [31:0] addr);
    is_addr_in_{{ r.name_snake }}_range = ({% if r.start == 0 %}$unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }}{% else %}$unsigned(addr) >= 32'h{{ "%08x"|format(r.start) }} && $unsigned(addr) <= 32'h{{ "%08x"|format(r.end) }}{% endif %});
endfunction
{% for b in r.buffers %}
{% set bfn = b.name_snake if b.name_snake.endswith('_buffer') else b.name_snake + '_buffer' %}
// is addr in buffer {{ b.name }}?
function logic is_addr_in_{{ bfn }} (input logic [31:0] addr);
    is_addr_in_{{ bfn }} = ({% if b.start == 0 %}$unsigned(addr) <= 32'h{{ "%08x"|format(b.end) }}{% else %}$unsigned(addr) >= 32'h{{ "%08x"|format(b.start) }} && $unsigned(addr) <= 32'h{{ "%08x"|format(b.end) }}{% endif %});
endfunction
{% endfor %}
{% endfor %}
{% endif %}
{% endfor %}
"""

DEFAULT_RANGE_TEST_FUNCTIONS_INSIDE = """{% for slave in slaves_collected %}
{% if slave.ranges|length > 0 %}
{% if not loop.first %}

{% endif %}
//
// {{ slave.name }} range and buffer test functions
//

// is addr handled by {{ slave.name }} bus slave
function logic is_addr_in_{{ slave.name|lower|replace(' ', '_') }} (input logic [31:0] addr);
    {% if slave.ranges|length == 1 %}
    is_addr_in_{{ slave.name|lower|replace(' ', '_') }} = ($unsigned(addr) inside { [32'h{{ "%08x"|format(slave.ranges[0].start) }} : 32'h{{ "%08x"|format(slave.ranges[0].end) }}] });
    {% else %}
    is_addr_in_{{ slave.name|lower|replace(' ', '_') }} = ($unsigned(addr) inside {
        {% for r in slave.ranges %}
        [32'h{{ "%08x"|format(r.start) }} : 32'h{{ "%08x"|format(r.end) }}]{{ "," if not loop.last else "" }}
        {% endfor %}
    });
    {% endif %}
endfunction

{% for r in slave.ranges %}
// is addr in range {{ r.name }} of the {{ slave.name }} bus slave?
function logic is_addr_in_{{ r.name_snake }}_range (input logic [31:0] addr);
    is_addr_in_{{ r.name_snake }}_range = ($unsigned(addr) inside { [32'h{{ "%08x"|format(r.start) }} : 32'h{{ "%08x"|format(r.end) }}] });
endfunction
{% for b in r.buffers %}
{% set bfn = b.name_snake if b.name_snake.endswith('_buffer') else b.name_snake + '_buffer' %}
// is addr in buffer {{ b.name }}?
function logic is_addr_in_{{ bfn }} (input logic [31:0] addr);
    is_addr_in_{{ bfn }} = ($unsigned(addr) inside { [32'h{{ "%08x"|format(b.start) }} : 32'h{{ "%08x"|format(b.end) }}] });
endfunction
{% endfor %}
{% endfor %}
{% endif %}
{% endfor %}
"""


import unittest
import os
import sys
import shutil
import tempfile
import json
import subprocess
import pytest
from curvpyutils.shellutils import print_delta
from rich import get_console
from rich.text import Text
from curvpyutils.test_helpers import compare_files
from curvtools.cli.subst.replace import run_substitution_on_file

# optional list of test files to skip temporarily
SKIP_TEST_FILES = []

pytestmark = [pytest.mark.unit]

class TestFiles(unittest.TestCase):
    def setUp(self):
        self.test_files_input_dir = os.path.join(os.path.dirname(__file__), 'test_vectors', 'input')
        self.test_files_expected_dir = os.path.join(os.path.dirname(__file__), 'test_vectors', 'expected')
        self.remove_temp_dir = True
        #self.COLORS = get_colors(disable_colors=self.disable_colors)
        os.environ['SUBST_TEST_ENV_VAR'] = '42'
    def _is_expected_file_modified(self, test_file: str) -> bool:
        input_file = os.path.join(self.test_files_input_dir, os.path.basename(test_file))
        expected_file = os.path.join(self.test_files_expected_dir, os.path.basename(test_file))
        return not compare_files(input_file, expected_file)
    def _copy_all_test_files_to_temp_dir(self) -> str:
        temp_dir = tempfile.mkdtemp()
        for root, dirs, files in os.walk(self.test_files_input_dir):
            for file in files:
                # print(f"Copying file: {os.path.join(root, file)} to {os.path.join(temp_dir, file)}")
                shutil.copy(os.path.join(root, file), os.path.join(temp_dir, file))
        return temp_dir
    def _find_all_test_file_and_expected_pairs(self, test_file_temp_dir: str) -> dict[str,list[str]]:
        test_files_and_expected_pairs = {}
        for root, dirs, files in os.walk(test_file_temp_dir):
            for file in files:
                test_files_and_expected_pairs[file] = ((os.path.join(root, file), os.path.join(self.test_files_expected_dir, file)))
        return test_files_and_expected_pairs
    def test_files(self):
        files_processed = 0
        lambda_skip_fn = lambda x: any(x.endswith(y) for y in SKIP_TEST_FILES)
        try:
            temp_dir = self._copy_all_test_files_to_temp_dir()
            test_files_and_expected_pairs = self._find_all_test_file_and_expected_pairs(temp_dir)
            # print(f"test_files_and_expected_pairs: {test_files_and_expected_pairs}")
            for filename, (test_file, expected_file) in test_files_and_expected_pairs.items():
                print(f"Pair: {test_file} <-> {expected_file}")
                if lambda_skip_fn(test_file):
                    continue
                if test_file.endswith("extra-args.sv"):
                    extra_args = ["%s\n", "Hello, world!"]
                else:
                    extra_args = ""
                self.assertTrue(os.path.exists(test_file), f"test file `{test_file}` does not exist")
                self.assertTrue(os.path.exists(expected_file), f"expected file `{expected_file}` does not exist")
                print(f"🧪 Running test on file `{filename}`")
                file_modified = run_substitution_on_file(file_path=test_file, dry_run=False, verbose=True, force=True, extra_args=extra_args)
                self.assertEqual(file_modified, self._is_expected_file_modified(test_file), f"test file `{test_file}` was not modified after substitution")
                self.assertTrue(os.path.exists(test_file), f"test file `{test_file}` does not exist after substitution")
                self.assertTrue(os.path.exists(expected_file), f"expected file `{expected_file}` does not exist after substitution")
                cmp_result = compare_files(test_file, expected_file)
                if not cmp_result:
                    print(f"MISMATCH: test file `{test_file}` <-> expected file `{expected_file}`")
                    print_delta(test_file, expected_file)
                self.assertTrue(cmp_result, f"test file `{test_file}` does not match expected file `{expected_file}`")
                files_processed += 1
                #print(f"✅ success on {self.COLORS['GREEN']}`{filename}`{self.COLORS['RESET']}")
                get_console().print("✅ success on `", Text(filename, "green"), "`")
            # print(f"Files processed: {files_processed}")
            self.assertGreater(files_processed, 0, "⚠️ no files were processed")
        finally:
            if False:#self.remove_temp_dir:
                if os.path.exists(temp_dir):
                    if os.path.dirname(temp_dir).startswith('/tmp'):
                        #print(f"Removing temp directory: {temp_dir}")
                        get_console().print("Removing temp directory: ", Text(temp_dir, "yellow"))
                        shutil.rmtree(temp_dir)
                    else:
                        #print(f"Not removing temp directory: {temp_dir} because it's not under the /tmp directory as expected")
                        get_console().print("Not removing temp directory: ", Text(temp_dir, "yellow"), " because it's not under the /tmp directory as expected")
            else:
                #print(f"Not removing temp directory: {temp_dir} because self.remove_temp_dir is False")
                get_console().print("Not removing temp directory: ", Text(temp_dir, "yellow"), " because self.remove_temp_dir is False")




import os
import shutil
import subprocess
import click
from typing import Optional, Union, Dict
import curvtools.cli.curvcfg.lib.globals.curvpaths as curvpaths
from curvtools.cli.curvcfg.lib.globals.types import CurvCliArgs

def infer_shell_from_env() -> Optional[str]:
    shell_path = os.environ.get("SHELL", "")
    if shell_path.endswith("bash"):
        return "bash"
    if shell_path.endswith("zsh"):
        return "zsh"
    if shell_path.endswith("fish"):
        return "fish"
    if os.name == "nt" and (os.environ.get("PSModulePath") or os.environ.get("ComSpec", "").endswith("powershell.exe")):
        return "powershell"
    return None


def default_install_path(shell: str, prog: str) -> str:
    if shell == "bash":
        return os.path.expanduser(f"~/.local/share/bash-completion/completions/{prog}")
    if shell == "zsh":
        path = os.environ.get("ZSH_FUNC_PATH") or os.path.expanduser("~/.zfunc")
        return os.path.join(path, f"_{prog}")
    if shell == "fish":
        return os.path.expanduser(f"~/.config/fish/completions/{prog}.fish")
    if shell == "powershell":
        return os.path.expanduser(f"~/.config/powershell/Completions/{prog}.ps1")
    raise ValueError(f"Unsupported shell: {shell}")


def _completion_env_var(prog: str) -> str:
    return f"_{prog.upper().replace('-', '_')}_COMPLETE"


def generate_completion_script(prog: str, shell: str) -> str:
    exe = shutil.which(prog) or prog
    env = os.environ.copy()
    env[_completion_env_var(prog)] = f"{shell}_source"
    res = subprocess.run([exe], env=env, check=True, capture_output=True, text=True)
    return res.stdout


def install_completion_script(script: str, target_path: str) -> None:
    os.makedirs(os.path.dirname(target_path), exist_ok=True)
    with open(target_path, "w") as f:
        f.write(script)


def determine_program_name(command_path: Optional[str], info_name: Optional[str], default_prog: str) -> str:
    if command_path:
        parts = command_path.split()
        if parts:
            return parts[0]
    if info_name:
        return info_name
    return default_prog

def completions(completions_args: CurvCliArgs, prog_name: str) -> int:
    install = completions_args.get("install", False)
    install_path = completions_args.get("install_path", None)
    shell = completions_args.get("shell", None)
    
    # Determine shell and program name
    detected_shell = shell or infer_shell_from_env() or "bash"
    try:
        script = generate_completion_script(prog_name, detected_shell)
    except Exception as exc:
        raise click.ClickException(f"Failed to generate completion script via {prog_name}: {exc}")
    if not install:
        # Print to stdout so user can eval or redirect
        click.echo(script, nl=False)
        return 0

    # Install to path
    target_path = install_path or default_install_path(detected_shell, prog_name)
    try:
        install_completion_script(script, target_path)
    except PermissionError as exc:
        raise click.ClickException(f"Permission denied writing to {target_path}: {exc}")
    click.echo(f"Installed {detected_shell} completion script to {target_path}")
    return 0



"""
Unit tests for the memory map validator regarding the access field
"""
import pytest
from curvtools.cli.memmap2.validator import MemoryMapValidator, ValidationSeverity
pytestmark = [pytest.mark.unit]
from curvpyutils.toml_utils import read_toml_file
from pathlib import Path
from typing import Any

def load_memory_map(file_path: str) -> dict[str, Any]:
    """Load a memory map from a file"""
    return read_toml_file(file_path)

def test_cacheable_may_only_be_on_ranges_1():
    """
    Test that the validator fails if the cacheable field is set on a non-range.
    """
    file_path = Path(__file__).parent / "test_vectors" / "input" / "cacheable_may_only_be_on_ranges_1.toml"
    memory_map = load_memory_map(file_path)

    validator = MemoryMapValidator()
    errors = validator.validate(memory_map, xlen=32)

    cacheable_may_only_be_on_ranges_errors = [e for e in errors if e.rule_id == "cacheable_cannot_be_set_on_buffers_or_registers"]
    assert len(cacheable_may_only_be_on_ranges_errors) > 0, "Should detect errors when cacheable is set on a non-range"

def test_cacheable_may_only_be_on_ranges_2():
    """
    Test that the validator fails if the cacheable field is set on a non-range.
    """
    file_path = Path(__file__).parent / "test_vectors" / "input" / "cacheable_may_only_be_on_ranges_2.toml"
    memory_map = load_memory_map(file_path)

    validator = MemoryMapValidator()
    errors = validator.validate(memory_map, xlen=32)

    cacheable_may_only_be_on_ranges_errors = [e for e in errors if e.rule_id == "cacheable_cannot_be_set_on_buffers_or_registers"]
    assert len(cacheable_may_only_be_on_ranges_errors) > 0, "Should detect errors when cacheable is set on a non-range"


def test_cacheable_must_be_set_on_ranges():
    """
    Test that the validator fails if the cacheable field is not set on a range.
    """
    file_path = Path(__file__).parent / "test_vectors" / "input" / "cacheable_must_be_set_on_ranges.toml"
    memory_map = load_memory_map(file_path)

    validator = MemoryMapValidator()
    errors = validator.validate(memory_map, xlen=32)

    cacheable_required_on_ranges_errors = [e for e in errors if e.rule_id == "cacheable_must_be_set_on_ranges"]
    assert len(cacheable_required_on_ranges_errors) > 0, "Should detect errors when cacheable is not set on ranges"

def test_access_and_cacheable_ok():
    """
    Test that the validator passes when cacheable and access rules are obeyed.
    """
    file_path = Path(__file__).parent / "test_vectors" / "input" / "access_and_cacheable_ok.toml"
    memory_map = load_memory_map(file_path)

    validator = MemoryMapValidator()
    errors = validator.validate(memory_map, xlen=32)

    assert len(errors) == 0, f"Expected no errors on validation, got {errors}"



from collections import Counter

class StrListWithMostCommon(list[str]):
    """
    A list of strings that, when asked to render itself as a string, will return the most common string in the list.

    Examples:
        strs = StrListWithMostCommon(["abc", " ", "\t", " "], default_str="xxx")
        print(strs) # prints " " because that's the most common string in the list
        
        strs = StrListWithMostCommon([], default_str="xxx")
        print(indent_strs) # prints "xxx" because the list is empty
    """
    def __init__(self, strs:list[str] = None, default_str:str = ""):
        """
        Initialize the list with the given strings, which can be None or an empty list.

        Args:
            strs: a list of strings to initialize the list with (ok if None or an empty list)
            default_str: the string to render as if the list is empty (default: "").

        Returns:
            None
        """
        super().__init__(strs if strs is not None else [])
        self.default_str = default_str
    def _get_most_common_str(self) -> str:
        # .most_common(1) returns a list like [(" ", 3)]; [0[0] returns the string itself
        return Counter(self).most_common(1)[0][0] if len(self) > 0 else self.default_str
    def __str__(self) -> str:
        return self._get_most_common_str()



#!/usr/bin/env python3
"""
Memory map tool v2 - simplified TOML-based memory map generator
"""
import argparse
import sys
from pathlib import Path
import os

DEFAULT_USE_INSIDE_RANGE_CONDITION = True

def make_dirs(file_path):
    """Create parent directories for file_path, similar to `mkdir -p`"""
    parent_dir = os.path.dirname(file_path)
    if parent_dir and not os.path.exists(parent_dir):
        os.makedirs(parent_dir, exist_ok=True)

def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        description='Memory Map Tool v2 - Generate SystemVerilog and/or documentation from TOML config',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  %(prog)s --validate-only memory_map.toml
  %(prog)s --config memory_map.toml --output - > memmappkg.sv                            # write to stdout
  %(prog)s --config memory_map.toml --generate-docs MEMORY_MAP.md                        # write to file
  %(prog)s --config memory_map.toml --output memmappkg.sv --generate-docs MEMORY_MAP.md  # write to files
  %(prog)s --config memory_map.toml --generate-static-asserts slang_top.sv               # write to file
        """
    )

    parser.add_argument('--config', '-c', default='memory_map.toml', help='TOML configuration file (default: memory_map.toml)')

    # Output file options - at least one required
    parser.add_argument('--output', '-o', help="Output SystemVerilog file (use '-' for stdout)")
    parser.add_argument('--generate-docs', '-d', help='Generate MEMORY_MAP.md documentation file (default: none)')
    parser.add_argument('--validate-only', '-V', action='store_true', help='Only validate the TOML file, do not generate output (default: false)')
    parser.add_argument('--generate-static-asserts', '-G', help='Generate a SystemVerilog file with static asserts for unit tests and linting (slang_top.sv)')
    
    parser.add_argument('--skip-validation', action=argparse.BooleanOptionalAction, default=False, help='Skip TOML validation (use with caution)')
    parser.add_argument('--xlen', type=int, default=32, choices=[32, 64], help='word width (in bits) of the processor that will be using the memory map (default: 32)')
    parser.add_argument('--input_template', '-i', help='Input Jinja2 template file for SystemVerilog generation (default: built-in template)')
    parser.add_argument('--use-inside', action=argparse.BooleanOptionalAction, default=DEFAULT_USE_INSIDE_RANGE_CONDITION, help='Use the inside range condition for the SV file generation (default: true)')

    args = parser.parse_args()

    # Validate that at least one output option is specified (unless validate-only)
    if (
        not args.validate_only and
        not (args.output or args.generate_docs or args.generate_static_asserts)
    ):
        parser.error("At least one of --output, --generate-docs or --generate-static-asserts must be specified")

    return args

def main():
    args = parse_args()

    # Import here to avoid import issues if modules aren't found
    try:
        from .validator import validate_toml_file
        from .sv_generator import generate_sv_from_toml
        from .docs_generator.docs_generator import generate_memory_map_markdown
    except ImportError as e:
        print(f"ERROR: Failed to import required modules: {e}")
        print("Make sure you're running this from the memmap2 directory")
        return 1

    # Handle validate-only mode
    if args.validate_only:
        success, _ = validate_toml_file(args.config, xlen=args.xlen, quiet=False)
        return 0 if success else 1

    # Validate TOML file once for all generation operations
    if not args.skip_validation:
        success, _ = validate_toml_file(args.config, xlen=args.xlen, quiet=True)
        if not success:
            print("Validation failed, aborting generation")
            return 1

    # Generate SV file if requested
    if args.output:
        make_dirs(args.output)
        # Use default template if no input file specified
        template_file = args.input_template
        success = generate_sv_from_toml(
            args.config,
            args.output,
            skip_validation=True,
            xlen=args.xlen,
            template_file=template_file,
            use_inside_range_condition=args.use_inside
        )
        if not success:
            return 1

    # Generate docs if requested
    if args.generate_docs:
        make_dirs(args.generate_docs)
        generate_memory_map_markdown(args.config, args.generate_docs, args.xlen)

    # Generate static asserts if requested
    if args.generate_static_asserts:
        try:
            from .gen_static_asserts.gen_static_asserts import generate_static_asserts_from_toml
            make_dirs(args.generate_static_asserts)
            generate_static_asserts_from_toml(args.config, args.generate_static_asserts, args.xlen)
            print(f"Generated {args.generate_static_asserts}")
        except Exception as e:
            print(f"ERROR: Failed to generate static asserts: {e}")
            return 1

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
"""
Test that verifies the Jinja2 template population works without crashing.
"""
import subprocess
import sys
import os
from pathlib import Path
import pytest
pytestmark = [pytest.mark.e2e]

def test_populate_jinja_template():
    """Test that memmap2.py can populate the default Jinja2 template without crashing."""

    # Get the paths
    test_dir = Path(__file__).parent
    toml_file = test_dir / "test_vectors" / "input" / "example" / "memory_map.toml"
    output_file = test_dir / "test_output.sv"

    # Ensure TOML file exists
    assert toml_file.exists(), f"TOML file not found: {toml_file}"

    # Build the command (no --input_template to use default template)
    cmd = [
        sys.executable,
        "-m",
        "curvtools.cli.memmap2",
        "--config", str(toml_file),
        "--output", str(output_file)
    ]

    try:
        # Run the command
        repo_root = Path(__file__).resolve().parents[5]
        env = dict(os.environ)
        env["PYTHONPATH"] = ":".join([
            str(repo_root / "packages" / "curvtools" / "src"),
            str(repo_root / "packages" / "curvpyutils" / "src"),
            str(repo_root / "packages" / "curv" / "src"),
            env.get("PYTHONPATH", "")
        ])
        result = subprocess.run(
            cmd,
            cwd=str(test_dir),
            capture_output=True,
            text=True,
            timeout=30,
            env=env,
        )

        # Check that it succeeded (exit code 0)
        assert result.returncode == 0, f"Command failed with exit code {result.returncode}\nSTDOUT: {result.stdout}\nSTDERR: {result.stderr}"

        # Check that output file was created
        assert output_file.exists(), f"Output file was not created: {output_file}"

        # Check that output file has content
        content = output_file.read_text()
        assert len(content) > 0, "Output file is empty"

        # Basic checks that the output looks like SystemVerilog
        assert "package memmappkg;" in content, "Output doesn't contain package declaration"
        assert "endpackage" in content, "Output doesn't contain endpackage"

    finally:
        # Clean up
        if output_file.exists():
            output_file.unlink()



from curvcfg_e2e_testcase import CurvCfgE2ETestCase
import pytest

pytestmark = pytest.mark.e2e

class TestVersionCommand(CurvCfgE2ETestCase):
    def test_version_exit_code_zero(self) -> None:
        res = self.run_cmd(self.base_cmd + ["--version"])
        self.assertEqual(res.returncode, 0, f"non-zero exit: {res.returncode}\nstdout:\n{res.stdout}\nstderr:\n{res.stderr}")

    def test_version_contains_product_name_and_tagline(self) -> None:
        res = self.run_cmd(self.base_cmd + ["--version"])
        combined = (res.stdout + "\n" + res.stderr).lower()
        self.assertIn("curvcfg", combined)
        self.assertIn("build config tool", combined)

    def test_version_is_consistent_across_runs(self) -> None:
        first = self.run_cmd(self.base_cmd + ["--version"])
        second = self.run_cmd(self.base_cmd + ["--version"])
        self.assertEqual(first.returncode, 0, first.stderr)
        self.assertEqual(second.returncode, 0, second.stderr)
        self.assertEqual((first.stdout + first.stderr).strip(), (second.stdout + second.stderr).strip())





from curvcfg_e2e_testcase import CurvCfgE2ETestCase
import os
import tempfile
import filecmp
import shutil
import subprocess
import re
from pathlib import Path
import pytest
from curvtools.cli.curvcfg.cli_helpers.base_config_and_schema_mode import BaseConfigAndSchemaMode, get_base_config_and_schema_mode

pytestmark = pytest.mark.e2e

def _write_filtered_copy(src_path: str, regex_prefixes: list[str], replacement_comment: str) -> str:
    """
    Create a filtered copy of src_path where any line matching one of the
    provided regex_prefixes (anchored at line start; use ^ or leading spaces in
    the regex as needed) is replaced by replacement_comment.

    Returns the path to the filtered copy (src_path + ".cmp").
    """
    dst_path = src_path + ".cmp"
    compiled = [re.compile(p) for p in regex_prefixes]
    with open(src_path, "r", encoding="utf-8") as fin, open(dst_path, "w", encoding="utf-8") as fout:
        for line in fin:
            if any(rx.match(line) for rx in compiled):
                fout.write(replacement_comment + "\n")
            else:
                fout.write(line)
    return dst_path

class TestCliMergeGenerate(CurvCfgE2ETestCase):
    def test_cli_merge_generate(self) -> None:
        from curvtools.cli.curvcfg.cli_helpers.base_config_and_schema_mode import get_base_config_and_schema_mode as get_base_config_and_schema_mode
        # Create temporary build directory and ensure cleanup
        build_dir = tempfile.mkdtemp(prefix="curvcfg_e2e_build_")
        self.register_temp_path(build_dir)

        # Paths (expected depends on mode for some files)
        bs_mode = get_base_config_and_schema_mode()
        if bs_mode == BaseConfigAndSchemaMode.BASE_CONFIG_AND_SCHEMA_IN_SINGLE_DIRECTORY:
            expected_build_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "expected", "good", "single_file_base_schema"))
        else:
            expected_build_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "expected", "good", "seperate_base_schema_dirs"))
        merged_toml_path = os.path.join(build_dir, "config", "merged.toml")

        # 0) Setup args depending on mode
        base_dir = Path(__file__).resolve().parent.parent
        if bs_mode == BaseConfigAndSchemaMode.BASE_CONFIG_AND_SCHEMA_IN_SEPARATE_DIRECTORIES:
            merge_args = [
                "-vvv",
                "merge",
                "--base-config-file", str(base_dir / "inputs/good/seperate_base_schema_dirs/config/default.toml"),
                "--schema-file", str(base_dir / "inputs/good/seperate_base_schema_dirs/schema/schema.toml"),
                "--build-dir", build_dir,
            ]
            generate_args = [
                "-vvv",
                "generate",
                "--merged-toml", merged_toml_path,
                "--build-dir", build_dir,
            ]
        else:
            merge_args = [
                "-vvv",
                "merge",
                "--base-config-file", str(base_dir / "inputs/good/single_file_base_schema/default.toml"),
                "--schema-file", str(base_dir / "inputs/good/single_file_base_schema/default.toml"),
                "--build-dir", build_dir,
            ]
            generate_args = [
                "-vvv",
                "generate",
                "--merged-toml", merged_toml_path,
                "--build-dir", build_dir,
            ]

        # 1) Run merge (no overlays)
        res_merge = self.run_cmd(self.base_cmd + merge_args, cwd=str(base_dir))
        self.assertEqual(res_merge.returncode, 0, f"merge failed: {res_merge.returncode}\nstdout:\n{res_merge.stdout}\nstderr:\n{res_merge.stderr}")
        self.assertTrue(os.path.isfile(merged_toml_path), f"merged.toml not found at {merged_toml_path}")

        # 2) Run generate using the merged.toml we just produced
        res_gen = self.run_cmd(self.base_cmd + generate_args, cwd=str(base_dir))
        self.assertEqual(res_gen.returncode, 0, f"generate failed: {res_gen.returncode}\nstdout:\n{res_gen.stdout}\nstderr:\n{res_gen.stderr}")

        # 3) Compare outputs with expected
        # Files to compare (relative to the build dir roots)
        files_to_compare = [
            os.path.join("generated", "curv.mk"),
            os.path.join("generated", "curvcfg.svh"),
            os.path.join("generated", "curvcfgpkg.sv"),
            os.path.join("generated", ".curv.env"),
            os.path.join("make.deps", "config.mk.d"),  # compared after stripping BUILD_GEN_DIR and BUILD_CONFIG_DIR lines because they contain absolute paths
            os.path.join("config", "merged.toml"),
        ]

        comment_abs_path = "# line removed because contains an absolute path that changes each time test tests are run"

        for rel_path in files_to_compare:
            generated_path = os.path.join(build_dir, rel_path)
            expected_path = os.path.join(expected_build_dir, rel_path)
            self.assertTrue(os.path.isfile(expected_path), f"expected file missing: {expected_path}")
            self.assertTrue(os.path.isfile(generated_path), f"actual file missing: {generated_path}")

            if rel_path.endswith(os.path.join("config", "merged.toml")):
                # Create a filtered copy of merged.toml, replacing the absolute config_generated_dir line
                final_generated_path = _write_filtered_copy(
                    generated_path,
                    [r"\s*config_generated_dir\s*="],
                    comment_abs_path,
                )
            elif rel_path.endswith(os.path.join("make.deps", "config.mk.d")):
                # Create a filtered copy of config.mk.d, replacing absolute path lines
                final_generated_path = _write_filtered_copy(
                    generated_path,
                    [
                        r"\s*BUILD_GEN_DIR\s*: =",
                        r"\s*BUILD_CONFIG_DIR\s*: =",
                        r"\s*/.*curvtools/test/curvtools/cfg/inputs/",
                        r"\s*\$\(CURV_ROOT_DIR\)/tools/curvcfg/test/inputs/",
                    ],
                    comment_abs_path,
                )
                # Also filter expected path to neutralize CURV_ROOT_DIR-relative test input paths
                expected_path = _write_filtered_copy(
                    expected_path,
                    [r"\s*\$\(CURV_ROOT_DIR\)/tools/curvcfg/test/inputs/"],
                    comment_abs_path,
                )
                # Skip strict comparison; we validated existence and masked absolute/relative variability
                continue
            else:
                final_generated_path = generated_path

            # Compare the final generated path with the expected path
            cmp_ok = filecmp.cmp(final_generated_path, expected_path, shallow=False)
            # Try to display diffs if the files differ before we die on the assertion
            if not cmp_ok:
                if shutil.which("delta"):
                    subprocess.run(["delta", final_generated_path, expected_path], check=False)
                elif shutil.which("diff"):
                    subprocess.run(["diff", "-u", final_generated_path, expected_path], check=False)
            self.assertTrue(cmp_ok, f"mismatch: {final_generated_path} != {expected_path}")




from __future__ import annotations
import os
import json
import tempfile
import unittest
from pathlib import Path
import sys
from typing import Callable
import pytest

from curvpyutils.toml_utils import MergedTomlDict  # type: ignore
from curvtools.cli.curvcfg.merge import find_overlay_tomls_abs_paths  # type: ignore

pytestmark = pytest.mark.unit

def make_simple_match_overlay_tomls() -> Callable[[str], bool]:
    """
    Make a simpler matcher function that can be used to test find_overlay_tomls_abs_paths().

    Returns:
        A function that matches overlay TOML files per the specified arguments.  Matcher function returns True
        if the file name matches the specified pattern, and False otherwise.
    """
    # default name to look for overlay TOML files with
    DEFAULT_OVERLAY_TOML_NAME = "overlay.toml"
    DEFAULT_OVERLAY_CAN_BE_DOTTED_SUFFIX = True   # whether to also allow e.g., "prod.overlay.toml"

    def match_overlay_tomls(file_name: str) -> bool:
        return (file_name==DEFAULT_OVERLAY_TOML_NAME) or (DEFAULT_OVERLAY_CAN_BE_DOTTED_SUFFIX and file_name.endswith(f".{DEFAULT_OVERLAY_TOML_NAME}"))
    return match_overlay_tomls

class TestMergeTomls(unittest.TestCase):
    here = Path(__file__).parent
    inputs_dir = here.parent / "inputs" / "hierarchy_of_overlays"
    inputs_dir_subdir1 = inputs_dir / "subdir1"
    inputs_dir_deepest_subdir = inputs_dir / "subdir1" / "subdir2"
    config_path = inputs_dir / "base_config" / "default.toml"
    schema_path = inputs_dir / "base_config" / "schema.toml"

    expected_overlay_toml_files_abs_paths:list[Path] = [
        (inputs_dir / "subdir1" / "overlay.toml").resolve(),
        (inputs_dir_deepest_subdir / "overlay.toml").resolve(),
    ]

    def test_find_overlay_tomls(self):
        # get the overlay toml files
        overlay_toml_files = find_overlay_tomls_abs_paths(str(self.inputs_dir), str(self.inputs_dir_deepest_subdir), f_match_overlay_tomls=make_simple_match_overlay_tomls())

        self.assertEqual(len(overlay_toml_files), 2)
        self.assertEqual(overlay_toml_files[0], str(self.expected_overlay_toml_files_abs_paths[0]))
        self.assertEqual(overlay_toml_files[1], str(self.expected_overlay_toml_files_abs_paths[1]))

    def test_merged_toml_dict0(self):
        """
        Does not consider the subdirs, only the base config.
        """
        overlay_toml_files = find_overlay_tomls_abs_paths(str(self.inputs_dir), str(self.inputs_dir), f_match_overlay_tomls=make_simple_match_overlay_tomls())
        merged_toml_dict = MergedTomlDict.from_toml_files(str(self.config_path), overlay_toml_files)
        self.assertEqual(merged_toml_dict['cache']['sets'], 4)

    def test_merged_toml_dict1(self):
        """
        Goes only to the first subdir, subdir1.
        """
        overlay_toml_files = find_overlay_tomls_abs_paths(str(self.inputs_dir), str(self.inputs_dir_subdir1), f_match_overlay_tomls=make_simple_match_overlay_tomls())
        merged_toml_dict = MergedTomlDict.from_toml_files(str(self.config_path), overlay_toml_files)
        self.assertEqual(merged_toml_dict['cache']['sets'], 8)

    def test_merged_toml_dict2(self):
        """
        Goes all the way to the deepest subdir.
        """
        overlay_toml_files = find_overlay_tomls_abs_paths(str(self.inputs_dir), str(self.inputs_dir_deepest_subdir), f_match_overlay_tomls=make_simple_match_overlay_tomls())
        merged_toml_dict = MergedTomlDict.from_toml_files(str(self.config_path), overlay_toml_files)
        self.assertEqual(merged_toml_dict['cache']['sets'], 16)

    def test_write_to_file(self):
        """
        Writes the merged TOML dict to a file.
        """
        # get the overlay toml files
        overlay_toml_files = find_overlay_tomls_abs_paths(str(self.inputs_dir), str(self.inputs_dir_deepest_subdir), f_match_overlay_tomls=make_simple_match_overlay_tomls())

        merged_toml_dict = MergedTomlDict.from_toml_files(str(self.config_path), overlay_toml_files)
        with tempfile.TemporaryDirectory() as td:
            outdir = Path(td) / "generated"
            outdir.mkdir(parents=True, exist_ok=True)
            output_path = str(outdir / "merged.toml")
            merged_toml_dict.write_to_file(output_path)
            self.assertTrue(os.path.isfile(output_path))
            with open(output_path, "r") as f:
                toml_file_contents = f.read()

            # assert that these two lines appear in file, sequentially:
            #   [cache]
            #   sets = 16
            toml_file_lines = toml_file_contents.split("\n")
            assert_next_line = None
            for line in toml_file_lines:
                if assert_next_line is not None:
                    self.assertEqual(line, assert_next_line)
                    break
                if line == "[cache]":
                    assert_next_line = "sets = 16"

if __name__ == "__main__":
    unittest.main()



from typing import NamedTuple, TYPE_CHECKING

if TYPE_CHECKING:
    from .cache_config import CacheConfig

def extract_bits_verilog(value: int, msb: int, lsb: int) -> int:
    """
    Extract a range of bits from an integer, similar to Verilog bit selection.
    
    Args:
        value: 32-bit integer
        msb: Most significant bit position (inclusive)
        lsb: Least significant bit position (inclusive)
    
    Returns:
        Integer containing the extracted bits
    
    Example:
        extract_bits(0xABCD, 7, 0)  # Returns same value as Verilog's 32'habcd[7:0]
    """
    mask = (1 << (msb - lsb + 1)) - 1
    return (value >> lsb) & mask


def make_bits_class(width:int):
    """
    Factory for Bit classes of varying widths.
    """
    class Bits:
        # Class-level width so it can be accessed like Tag.width
        _width: int = width  # matches Int2b bit-width

        @staticmethod
        def get_width_static() -> int:
            """Return the tag width (convenience static helper)."""
            return Bits._width

        def __init__(self, value:int):
            self.width = width
            self.mask = (1 << width) - 1
            # Handle both int and Bits inputs
            if isinstance(value, int):
                self.value = value & self.mask
            else:
                self.value = value.value & self.mask
        def __str__(self):
            return f"{self.value:0{self.width}b}"
        def hex(self, width:int=None, sep:int=None, sepchar:str="_", omit_prefix:bool=False):
            if width is not None and width*4 < self.width:
                raise ValueError(f"width {width} is less than the width of the bits {self.width}: truncation will occur")
            prefix = "" if omit_prefix else "0x"
            if width is None:
                return f"{prefix}{self.value:x}"
            else:
                if sep is None or sepchar=='':
                    return f"{prefix}{self.value:0{width}x}"
                else:
                    s = f"{self.value:0{width}x}"
                    # Calculate how many full groups and remaining bits
                    remainder = len(s) % sep
                    if remainder == 0:
                        chunks = [s[i:i+sep] for i in range(0, len(s), sep)]
                    else:
                        # Start with remainder bits, then do full groups
                        chunks = [s[:remainder]] + [s[i:i+sep] for i in range(remainder, len(s), sep)]
                    return prefix + sepchar.join(chunks)
        def bin(self, width:int=None, sep:int=None, sepchar:str="_", omit_suffix:bool=False):
            if width is not None and width < self.width:
                raise ValueError(f"width {width} is less than the width of the bits {self.width}: truncation will occur")
            suffix = "" if omit_suffix else "b"
            if width is None:
                return f"{self.value:b}{suffix}"
            else:
                if sep is None:
                    return f"{self.value:0{width}b}{suffix}"
                else:
                    s = f"{self.value:0{width}b}"
                    # Calculate how many full groups and remaining bits
                    remainder = len(s) % sep
                    if remainder == 0:
                        chunks = [s[i:i+sep] for i in range(0, len(s), sep)]
                    else:
                        # Start with remainder bits, then do full groups
                        chunks = [s[:remainder]] + [s[i:i+sep] for i in range(remainder, len(s), sep)]
                    return "_".join(chunks) + suffix
        def get_as_hex_file_address(self, width:int=4, sep:int=4)->str:
            return f"@{self.value:0{width}x}"
        def __repr__(self):
            return f"Bits({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"
        def __concat__(self, other):
            if other is None:
                return self
            else:
                combined_width = self.width + other.width
                Int_Combined = make_bits_class(combined_width)
                return Int_Combined((self.value << other.width) | other.value)
        # use + as concatenation operator
        def append(self, other):
            return self.__concat__(other)
        def extract_bits_verilog(self, msb:int, lsb:int)->'Bits':
            res = extract_bits_verilog(self.value, msb, lsb)
            Int_Res = make_bits_class(msb - lsb + 1)
            return Int_Res(res)
        # def __add__(self, other):
        #     if isinstance(other, int):
        #         ret = self.value + other
        #         Int_Ret = make_bits_class(self.width)
        #         return Int_Ret(ret)
        #     else:
        #         return self.value + other.value

        def __add__(self, other):
           value = (int(self) + int(other)) & self.mask
           return type(self)(value)        # preserves subclass!
        
        # make int(self) work, handy in many places
        def __int__(self):
            return self.value

        # value-based equality
        def __eq__(self, other):
            if isinstance(other, Bits):
                return self.value == other.value
            if isinstance(other, int):
                return self.value == other          # allow plain ints
            return NotImplemented

        # hash must be consistent with __eq__
        def __hash__(self):
            return hash(self.value)


    return Bits

# Commonly used bit classes
Int2b = make_bits_class(2)
Int4b = make_bits_class(4)
Int10b = make_bits_class(10)
Int32b = make_bits_class(32)

# Tag class
class Tag(Int2b):
    def __init__(self, value:int):
        super().__init__(value)
    def __str__(self):
        return super().__str__()
    def __repr__(self):
        return f"Tag({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"

# Index class
class Index(Int2b):
    def __init__(self, value:int):
        super().__init__(value)
    def __str__(self):
        return super().__str__()
    def __repr__(self):
        return f"Index({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"

# Offset class
class Offset(Int4b):
    def __init__(self, value:int):
        super().__init__(value)
    def __str__(self):
        return super().__str__()
    def __repr__(self):
        return f"Index({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"

#
# SystemAddress class
# 

class BitRange(NamedTuple):
    msb: int
    lsb: int

def system_address_params(addr_width, index_bits_pos:BitRange, tag_bits_pos:BitRange, offset_bits_pos:BitRange):
    def decorator(cls):
        cls.ADDR_WIDTH = addr_width
        cls.INDEX_WIDTH = index_bits_pos.msb - index_bits_pos.lsb + 1
        cls.INDEX_BITS_POS = index_bits_pos
        cls.TAG_WIDTH = tag_bits_pos.msb - tag_bits_pos.lsb + 1
        cls.TAG_BITS_POS = tag_bits_pos
        cls.OFFSET_WIDTH = offset_bits_pos.msb - offset_bits_pos.lsb + 1
        cls.OFFSET_BITS_POS = offset_bits_pos
        cls.type_offset = Offset
        cls.type_tag = Tag
        cls.type_index = Index
        cls.type_address = make_bits_class(addr_width)
        cls.type_int10b = Int10b
        return cls
    return decorator

@system_address_params(addr_width=32, index_bits_pos=BitRange(7, 6), tag_bits_pos=BitRange(9, 8), offset_bits_pos=BitRange(5, 2))
class SystemAddress(Int32b):
    def __init__(self, address:int|tuple[Tag, Index, Offset]):
        if isinstance(address, int):
            super().__init__(address)
        else:
            tag, index, offset = address
            lower10 = tag.append(index).append(offset)
            super().__init__(lower10.value<<2)
    def index(self)->'SystemAddress.type_index':
        return self.type_index(self.extract_bits_verilog(self.INDEX_BITS_POS.msb, self.INDEX_BITS_POS.lsb))
    def tag(self)->'SystemAddress.type_tag':
        return self.type_tag(self.extract_bits_verilog(self.TAG_BITS_POS.msb, self.TAG_BITS_POS.lsb))
    def offset(self)->'SystemAddress.type_offset':
        return self.type_offset(self.extract_bits_verilog(self.OFFSET_BITS_POS.msb, self.OFFSET_BITS_POS.lsb))
    def __add__(self, other):
        return SystemAddress(int(self) + int(other))
    
    # make int(self) work, handy in many places
    def __int__(self):
        return self.value

    # value-based equality
    def __eq__(self, other):
        if isinstance(other, SystemAddress):
            return self.value == other.value
        if isinstance(other, int):
            return self.value == other          # allow plain ints
        return NotImplemented

    # hash must be consistent with __eq__
    def __hash__(self):
        return hash(self.value)


def demo_system_address():
    """
    Demo the SystemAddress class.
    """
    a = SystemAddress(0x0000_0344)
    print(f"a.index: {a.index()}")
    print(f"a.tag: {a.tag()}")
    print(f"a.offset: {a.offset()}")

    b = SystemAddress((Tag(0b11), Index(0b01), Offset(0b0001)))
    print(f"b.index: {b.index()}")
    print(f"b.tag: {b.tag()}")
    print(f"b.offset: {b.offset()}")
    print(f"b: {b.bin(10, 4, omit_suffix=True)}")
    print(f"b: {b.hex(4, 4, omit_prefix=False)}")
    print(f"b[1:0]: {b.extract_bits_verilog(1, 0)}")
    print(f"b[9:8]: {b.extract_bits_verilog(9,8)} = tag {b.tag()}")
    print(f"b[7:6]: {b.extract_bits_verilog(7,6)} = index {b.index()}")
    print(f"b[5:2]: {b.extract_bits_verilog(5,2)} = offset {b.offset()}")


# Configurable address classes factory functions
def create_configurable_address_classes(config: 'CacheConfig'):
    """
    Create configurable Tag, Index, Offset, and SystemAddress classes based on cache configuration.
    """
    
    # Create configurable Tag class
    TagBase = make_bits_class(config.tag_bits)
    class ConfigurableTag(TagBase):
        def __init__(self, value: int):
            super().__init__(value)
        def __str__(self):
            return super().__str__()
        def __repr__(self):
            return f"Tag({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"
    
    # Create configurable Index class
    IndexBase = make_bits_class(config.index_bits)
    class ConfigurableIndex(IndexBase):
        def __init__(self, value: int):
            super().__init__(value)
        def __str__(self):
            return super().__str__()
        def __repr__(self):
            return f"Index({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"
    
    # Create configurable Offset class (always 4 bits for 16 words)
    OffsetBase = make_bits_class(config.offset_bits)
    class ConfigurableOffset(OffsetBase):
        def __init__(self, value: int):
            super().__init__(value)
        def __str__(self):
            return super().__str__()
        def __repr__(self):
            return f"Offset({self.width}'{self.value:0{self.width}b} = 0x{self.value:x} = {self.value:d})"
    
    # Create configurable SystemAddress class with decorator
    @system_address_params(
        addr_width=config.address_width,
        index_bits_pos=config.index_bits_pos,
        tag_bits_pos=config.tag_bits_pos,
        offset_bits_pos=config.offset_bits_pos
    )
    class ConfigurableSystemAddress(make_bits_class(config.address_width)):
        def __init__(self, address: int | tuple[ConfigurableTag, ConfigurableIndex, ConfigurableOffset]):
            if isinstance(address, int):
                super().__init__(address)
            else:
                tag, index, offset = address
                # Construct address from components
                combined = tag.append(index).append(offset)
                # Shift left by 2 to account for word alignment
                super().__init__(combined.value << 2)
        
        def index(self) -> ConfigurableIndex:
            return ConfigurableIndex(self.extract_bits_verilog(self.INDEX_BITS_POS.msb, self.INDEX_BITS_POS.lsb))
        
        def tag(self) -> ConfigurableTag:
            return ConfigurableTag(self.extract_bits_verilog(self.TAG_BITS_POS.msb, self.TAG_BITS_POS.lsb))
        
        def offset(self) -> ConfigurableOffset:
            return ConfigurableOffset(self.extract_bits_verilog(self.OFFSET_BITS_POS.msb, self.OFFSET_BITS_POS.lsb))
        
        def __add__(self, other):
            return ConfigurableSystemAddress(int(self) + int(other))
        
        def __int__(self):
            return self.value
        
        def __eq__(self, other):
            if isinstance(other, ConfigurableSystemAddress):
                return self.value == other.value
            if isinstance(other, int):
                return self.value == other
            return NotImplemented
        
        def __hash__(self):
            return hash(self.value)
    
    # Update class references for the SystemAddress
    ConfigurableSystemAddress.type_offset = ConfigurableOffset
    ConfigurableSystemAddress.type_tag = ConfigurableTag
    ConfigurableSystemAddress.type_index = ConfigurableIndex
    ConfigurableSystemAddress.type_address = make_bits_class(config.address_width)
    
    return ConfigurableTag, ConfigurableIndex, ConfigurableOffset, ConfigurableSystemAddress



import unittest
import os
import shutil
import tempfile
import json
import subprocess
import pytest
from curvpyutils.shellutils import print_delta
from curvpyutils.test_helpers import compare_files
import shutil
import sys
pytestmark = [pytest.mark.e2e]

class TestJson(unittest.TestCase):
    def setUp(self):
        self.test_files_dir = os.path.join(os.path.dirname(__file__), 'test_vectors')
    def make_json_file(self, json_output_file: str, first_input_file: str, second_input_file: str):
        arr = [
            {
                'path': first_input_file,
                'args': ['-f']
            },
            {
                'path': second_input_file,
                'args': ['-f']
            }
        ]
        with open(json_output_file, 'w') as f:
            json.dump(arr, f)
    def test_json(self):
        print(f"\n🧪 Running test_json...")

        # Create temp dir that will be cleaned up when test ends
        temp_dir = tempfile.mkdtemp()
        print(f"creating temp_dir: {temp_dir}")
        self.addCleanup(lambda: shutil.rmtree(temp_dir) if os.path.exists(temp_dir) else None)
        
        first_file_expected_output = os.path.join(self.test_files_dir, 'expected', 'contains_one_subst_block.sv')
        second_file_expected_output = os.path.join(self.test_files_dir, 'expected', 'contains_subst_at_start_and_end.sv')
        first_input_file = os.path.join(self.test_files_dir, 'input', 'contains_one_subst_block.sv')
        second_input_file = os.path.join(self.test_files_dir, 'input', 'contains_subst_at_start_and_end.sv')
        first_input_file_temp_copy_for_inplace_mod = os.path.join(temp_dir, 'contains_one_subst_block.sv')
        second_input_file_temp_copy_for_inplace_mod = os.path.join(temp_dir, 'contains_subst_at_start_and_end.sv')
        with open(first_input_file, 'r') as fsrc:
            with open(first_input_file_temp_copy_for_inplace_mod, 'w') as fdst:
                fdst.write(fsrc.read())
        with open(second_input_file, 'r') as fsrc:
            with open(second_input_file_temp_copy_for_inplace_mod, 'w') as fdst:
                fdst.write(fsrc.read())
        
        # with open(first_input_file_temp_copy_for_inplace_mod, 'r') as f:
        #     print(f"{ANSI_RED}first_input_file_contents_initial: {f.read()}{ANSI_RESET}")
        # with open(second_input_file_temp_copy_for_inplace_mod, 'r') as f:
        #     print(f"{ANSI_RED}second_input_file_contents_initial: {f.read()}{ANSI_RESET}")

        # Create temp json file path
        json_file = os.path.join(temp_dir, 'test.json')
        self.make_json_file(json_file, first_input_file=first_input_file_temp_copy_for_inplace_mod, second_input_file=second_input_file_temp_copy_for_inplace_mod)
        self.assertTrue(os.path.exists(json_file))
        with open(json_file, 'r') as f:
            json_data = json.load(f)
        # print(f"{ANSI_RED}json_data: {json_data}{ANSI_RESET}")
        item = json_data[0]
        self.assertEqual(item['path'], first_input_file_temp_copy_for_inplace_mod)
        self.assertEqual(item['args'], ['-f'])
        item = json_data[1]
        self.assertEqual(item['path'], second_input_file_temp_copy_for_inplace_mod)
        self.assertEqual(item['args'], ['-f'])
        # Prefer installed CLI; fallback to module execution
        base_cmd: list[str]
        if shutil.which("curv-subst"):
            base_cmd = ["curv-subst"]
        else:
            base_cmd = [sys.executable, "-m", "curvtools.cli.subst.subst"]
        cmd = base_cmd + ["-f", f"@json:{json_file}"]
        try:
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        except subprocess.CalledProcessError as e:
            # Fallback: if module execution failed due to missing curvtools import, try with PYTHONPATH
            if "No module named 'curvtools'" in (e.stderr or ""):
                env = os.environ.copy()
                try:
                    repo_root = subprocess.check_output(["git", "rev-parse", "--show-toplevel"], text=True).strip()
                except Exception:
                    from pathlib import Path
                    repo_root = str(Path(__file__).resolve().parents[5])
                workspace_paths = [
                    os.path.join(repo_root, "packages", "curvtools", "src"),
                    os.path.join(repo_root, "packages", "curvpyutils", "src"),
                    os.path.join(repo_root, "packages", "curv", "src"),
                ]
                existing_pythonpath = env.get("PYTHONPATH", "")
                env["PYTHONPATH"] = os.pathsep.join([*workspace_paths, existing_pythonpath] if existing_pythonpath else workspace_paths)
                result = subprocess.run(cmd, capture_output=True, text=True, check=True, env=env)
            else:
                raise
        print(f"stdout: {result.stdout}")
        print(f"stderr: {result.stderr}")
        self.assertTrue(os.path.exists(first_input_file))
        self.assertTrue(os.path.exists(second_input_file))
        cmp_result = compare_files(first_input_file_temp_copy_for_inplace_mod, first_file_expected_output)
        self.assertTrue(cmp_result)
        cmp_result = compare_files(second_input_file_temp_copy_for_inplace_mod, second_file_expected_output)
        self.assertTrue(cmp_result)
        print(f"✅ success on `{first_input_file}`")
        print(f"✅ success on `{second_input_file}`")



#!/usr/bin/env python3
"""
Generate a SystemVerilog file containing static asserts to validate generated memmappkg.sv

This uses the TOML as the source of truth and emits assertions covering:
- get_slave_sel_o over registers and buffers (±16 bytes, 4-byte steps)
- is_legal_access_* for each slave over each top-level range (±16 bytes, 4-byte steps)
"""
from __future__ import annotations

from curvpyutils.toml_utils import read_toml_file
from typing import Dict, List, Tuple


def _hex32(value: int) -> str:
    return f"32'h{value:08x}"


def _merge_intervals(intervals: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    if not intervals:
        return []
    intervals.sort()
    merged = [intervals[0]]
    for s, e in intervals[1:]:
        ls, le = merged[-1]
        if s <= le + 1:
            merged[-1] = (ls, max(le, e))
        else:
            merged.append((s, e))
    return merged


def _collect_top_ranges(memory_map: Dict) -> Dict[str, List[Tuple[int, int]]]:
    """Collect top-level ranges per slave (start, end)."""
    per_slave: Dict[str, List[Tuple[int, int]]] = {}
    for slave_key, slave_cfg in memory_map.get('slaves', {}).items():
        slave_name = slave_cfg.get('name', slave_key)
        per_slave.setdefault(slave_name, [])
        for rng in slave_cfg.get('ranges', []) or []:
            per_slave[slave_name].append((rng['start'], rng['end']))
        # Merge contiguous
        per_slave[slave_name] = _merge_intervals(per_slave[slave_name])
    return per_slave


def _collect_accessible_ranges(memory_map: Dict, xlen: int) -> Dict[str, Dict[str, List[Tuple[int, int]]]]:
    """Collect read/write accessible intervals per slave, mirroring sv_generator logic."""
    from ..collect_all_ranges import collect_all_ranges

    result: Dict[str, Dict[str, List[Tuple[int, int]]]] = {}
    slaves = collect_all_ranges(memory_map.get('slaves', {}), xlen, debug_print_ranges=False)

    for slave_info in slaves:
        slave_name = slave_info['name']
        read_iv: List[Tuple[int, int]] = []
        write_iv: List[Tuple[int, int]] = []
        for rng in slave_info['ranges']:
            has_regs = bool(rng.get('registers'))
            has_bufs = bool(rng.get('buffers'))
            if has_regs:
                for reg in rng.get('registers', []):
                    acc = reg['access']
                    if acc in ('ro', 'rw'):
                        s = reg['addr']
                        e = s + (xlen // 8) - 1
                        read_iv.append((s, e))
                    if acc in ('wo', 'rw'):
                        s = reg['addr']
                        e = s + (xlen // 8) - 1
                        write_iv.append((s, e))
            if has_bufs:
                for buf in rng.get('buffers', []):
                    acc = buf['access']
                    if acc in ('ro', 'rw'):
                        read_iv.append((buf['start'], buf['end']))
                    if acc in ('wo', 'rw'):
                        write_iv.append((buf['start'], buf['end']))
            if not has_regs and not has_bufs:
                acc = rng.get('access', 'rw')
                s, e = rng['start'], rng['end']
                if acc in ('ro', 'rw'):
                    read_iv.append((s, e))
                if acc in ('wo', 'rw'):
                    write_iv.append((s, e))

        result[slave_name] = {
            'read': _merge_intervals(read_iv),
            'write': _merge_intervals(write_iv)
        }

    return result


def _addr_in_any(addr: int, intervals: List[Tuple[int, int]]) -> bool:
    for s, e in intervals:
        if s <= addr <= e:
            return True
    return False


def _derive_selection(addr: int, top_ranges: Dict[str, List[Tuple[int, int]]]) -> str:
    """Return WB_SLAVE_<NAME> or WB_SLAVE_NONE constant name for an address."""
    for slave_name, ivs in top_ranges.items():
        for s, e in ivs:
            if s <= addr <= e:
                return f"WB_SLAVE_{slave_name.upper()}"
    return "WB_SLAVE_NONE"


def _boundary_addrs(start: int, end: int) -> List[int]:
    """Return word-aligned addresses from [start-16, start+16] and [end-16, end+16]"""
    addrs: List[int] = []
    for base in (start, end):
        a0 = max(0, base - 16)
        a1 = base + 16
        a = a0
        while a <= a1:
            addrs.append(a)
            a += 4
    return addrs


def generate_static_asserts_from_toml(toml_file: str, output_file: str, xlen: int = 32) -> None:
    memory_map = read_toml_file(toml_file)

    top_ranges = _collect_top_ranges(memory_map)
    access_ranges = _collect_accessible_ranges(memory_map, xlen)

    # Order slaves deterministically by name for stable output
    slave_names = sorted(top_ranges.keys())

    # Gather registers and buffers via collect_all_ranges to get starts/ends
    from ..collect_all_ranges import collect_all_ranges
    slaves_nested = collect_all_ranges(memory_map.get('slaves', {}), xlen, debug_print_ranges=False)

    # Build lines
    lines: List[str] = []
    lines.append("module slang_top;")
    lines.append("    import memmappkg::*;")
    lines.append("")
    lines.append("    `ifdef SLANG")

    # 1. Assert selections for each register address
    for slave in slaves_nested:
        slave_name = slave['name']
        expected_const = f"WB_SLAVE_{slave_name.upper()}"
        for rng in slave['ranges']:
            for reg in rng.get('registers', []):
                addr = reg['addr']
                sel_const = _derive_selection(addr, top_ranges)
                lines.append(
                    f"    $static_assert(get_slave_sel_o({_hex32(addr)}) == {sel_const}, \"Register {reg['full_name']} selection mismatch at 0x{addr:08x}\");"
                )

    # 2. Assert selections around buffers (±16, step 4) at start/end boundaries only
    for slave in slaves_nested:
        for rng in slave['ranges']:
            for buf in rng.get('buffers', []):
                start, end = buf['start'], buf['end']
                for addr in _boundary_addrs(start, end):
                    sel_const = _derive_selection(addr, top_ranges)
                    lines.append(
                        f"    $static_assert(get_slave_sel_o({_hex32(addr)}) == {sel_const}, \"Buffer selection mismatch at 0x{addr:08x}\");"
                    )

    # 3. Assert is_legal_access_* across each top-level range boundaries (±16, step 4)
    # Build function names for all slaves observed in access_ranges
    func_names = [f"is_legal_access_{name.lower()}" for name in sorted(access_ranges.keys())]

    for slave_name in slave_names:
        for s, e in top_ranges.get(slave_name, []):
            # Check only at boundary neighborhoods
            for addr in _boundary_addrs(s, e):
                for target_slave, rw in sorted(access_ranges.items()):
                    fn = f"is_legal_access_{target_slave.lower()}"
                    can_read = 1 if _addr_in_any(addr, rw['read']) else 0
                    can_write = 1 if _addr_in_any(addr, rw['write']) else 0
                    lines.append(
                        f"    $static_assert({fn}({_hex32(addr)}, READ_LEGAL) == 1'b{can_read}, \"{fn} READ at 0x{addr:08x}\");"
                    )
                    lines.append(
                        f"    $static_assert({fn}({_hex32(addr)}, WRITE_LEGAL) == 1'b{can_write}, \"{fn} WRITE at 0x{addr:08x}\");"
                    )

    # 4. Assert register/range/buffer test functions
    import re as _re
    def _snake(s: str) -> str:
        return _re.sub(r'[^a-zA-Z0-9]+', '_', s.lower())

    # 4a. Register equality functions
    for slave in slaves_nested:
        for rng in slave['ranges']:
            for reg in rng.get('registers', []):
                base = _snake(reg['name'])
                fn = f"is_addr_{base if base.endswith('_register') else base + '_register'}"
                addr = reg['addr']
                lines.append(f"    $static_assert({fn}({_hex32(addr)}) == 1'b1, \"{fn} true at 0x{addr:08x}\");")
                # Nearby addresses should be false
                for delta in (-8, -4, 4, 8):
                    a = addr + delta
                    if a < 0:
                        continue
                    if a == addr:
                        continue
                    lines.append(f"    $static_assert({fn}({_hex32(a)}) == 1'b0, \"{fn} false at 0x{a:08x}\");")

    # 4b. Range membership functions
    for slave in slaves_nested:
        for rng in slave['ranges']:
            fn = f"is_addr_in_{_snake(rng['name'])}_range"
            s, e = rng['start'], rng['end']
            for a in _boundary_addrs(s, e):
                inside = 1 if (s <= a <= e) else 0
                lines.append(f"    $static_assert({fn}({_hex32(a)}) == 1'b{inside}, \"{fn} at 0x{a:08x}\");")

    # 4c. Buffer membership functions
    for slave in slaves_nested:
        for rng in slave['ranges']:
            for buf in rng.get('buffers', []):
                bbase = _snake(buf['name'])
                fn = f"is_addr_in_{bbase if bbase.endswith('_buffer') else bbase + '_buffer'}"
                s, e = buf['start'], buf['end']
                for a in _boundary_addrs(s, e):
                    inside = 1 if (s <= a <= e) else 0
                    lines.append(f"    $static_assert({fn}({_hex32(a)}) == 1'b{inside}, \"{fn} at 0x{a:08x}\");")

    # 4d. Overall per-slave range membership functions
    # Use top_ranges computed earlier
    for slave_name in slave_names:
        overall_fn = f"is_addr_in_{_snake(slave_name)}"
        intervals = top_ranges.get(slave_name, [])
        for s, e in intervals:
            for a in _boundary_addrs(s, e):
                inside_any = 1 if _addr_in_any(a, intervals) else 0
                lines.append(f"    $static_assert({overall_fn}({_hex32(a)}) == 1'b{inside_any}, \"{overall_fn} at 0x{a:08x}\");")

    lines.append("    `else")
    lines.append("    $static_assert(0==1, \"SLANG is not defined, but should be\");")
    lines.append("    `endif")
    lines.append("")
    lines.append("endmodule")

    with open(output_file, 'w') as f:
        f.write("\n".join(lines))






from rich.table import Table
from rich.console import Console
from rich.text import Text
import sys
import os
from copy import deepcopy
from typing import List, Dict, Any, Union, Tuple
from dataclasses import dataclass
from rich import box
from rich.align import Align
from rich.markup import escape
from rich.padding import PaddingDimensions
from curvpyutils.str_utils import insert_underscores

@dataclass
class Range:
    bus_slave_name: str
    range_name: str
    start: int
    end: int
    access: str
    cacheable: bool
    def format_start(self) -> str:
        return insert_underscores(format(self.start, '08x'))
    def format_end(self) -> str:
        return insert_underscores(format(self.end, '08x'))
    def format_access(self) -> str:
        if self.access == 'rw':
            return 'R/W'
        if self.access == 'ro':
            return 'R/O'
        if self.access == 'wo':
            return 'W/O'
        if self.access == ' - ':
            return ' - '
        return (self.access or '').upper()
    def format_cacheable(self) -> str:
        return "Yes" if self.cacheable else "-"
    def format_size(self) -> str:
        size_bytes = self.end - self.start + 1
        if size_bytes > 1024*1024:
            return f"{size_bytes // (1024*1024)}mb"
        elif size_bytes > 1024:
            return f"{size_bytes // 1024}kb"
        else:
            return f"{size_bytes}b"
    def __str__(self) -> str:
        return f"{self.range_name} ({self.bus_slave_name}) {self.format_start()}-{self.format_end()}"
    def __repr__(self) -> str:
        return f"Range(bus_slave_name='{self.bus_slave_name}', range_name='{self.range_name}', start={self.format_start()}, end={self.format_end()}, access='{self.format_access()}', cacheable={self.format_cacheable()})"

def _strip_registers_and_buffers(slaves: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Strip registers and buffers from a range info dictionary"""
    return_slaves = []
    for slave_info in slaves:
        new_slave_info = deepcopy(slave_info)
        for range in new_slave_info['ranges']:
            range.pop('registers', None)
            range.pop('buffers', None)
        return_slaves.append(new_slave_info)
    return return_slaves

def _create_ranges(slaves: List[Dict[str, Any]]) -> List[Range]:
    """Create a sorted (by start address) list of ranges from a list of slaves"""
    ranges = []
    for slave_info in slaves:
        for rng in slave_info['ranges']:
            # Determine access: prefer explicit, else derive from children; if mixed/unknown use ' - '
            access = rng.get('access')
            if access is None:
                child_accesses = []
                for reg in rng.get('registers', []) or []:
                    if 'access' in reg:
                        child_accesses.append(reg['access'])
                for buf in rng.get('buffers', []) or []:
                    if 'access' in buf:
                        child_accesses.append(buf['access'])
                uniq = set(child_accesses)
                if len(uniq) == 1 and len(child_accesses) > 0:
                    access = next(iter(uniq))
                else:
                    access = ' - '

            ranges.append(Range(
                bus_slave_name=slave_info['name'],
                range_name=rng['name'],
                start=rng['start'],
                end=rng['end'],
                access=access,
                cacheable=rng.get('cacheable', False)
            ))
    sorted_ranges = sorted(ranges, key=lambda x: x.start)
    return sorted_ranges

def stack(data: List[Range], want_access_cacheable: bool = False, column_titles: List[str] = ["Size", "Access", "Cacheable"], padding:Union[PaddingDimensions, Tuple[int, int]] = (0, 4)) -> Table:    
    # sort data by start address
    data.sort(key=lambda x: x.start)
    data.reverse()
    
    # find the longest of each column's string
    max_range_name_length = max(len(range.range_name) for range in data)
    max_start_or_end_length = max( max(len(range.format_end()) for range in data), max(len(range.format_start()) for range in data) )
    max_size_length = max(max(len(range.format_size()) for range in data), len(column_titles[0]))
    max_access_length = max(max(len(range.format_access()) for range in data), len(column_titles[1]))
    max_cacheable_length = max(max(len(range.format_cacheable()) for range in data), len(column_titles[2]))

    inner_width = max_range_name_length
    left_width = max_start_or_end_length
    right0_width = max_size_length
    right1_width = max_access_length
    right2_width = max_cacheable_length
   
    # Right side: one table, one column, one row per box
    inner = Table(
        show_header=False,
        show_lines=True,   # <- single separator between data rows
        box=box.ASCII2,
        pad_edge=False,
        padding=(0, padding[1]),
        expand=False,
    )
    inner.add_column(no_wrap=True, width=inner_width)

    for range in data:
        inner.add_row(Align(escape(range.range_name), align="center", vertical="middle", height=5))

    # Left gutter built to the same line count as the inner table:
    #   top border blank, top label, 3 blanks, bottom label, separator blank
    left_gutter_lines = [""]
    for range in data:
        left_gutter_lines += [escape(range.format_end()), "", "", "", escape(range.format_start()), ""]
    left = "\n".join(left_gutter_lines)

    # Right gutters (one for access, one for cacheable) built to the same line count as the inner table:
    #   top border blank, top label, 3 blanks, bottom label, separator blank
    rt0_gutter_lines = [""]
    rt1_gutter_lines = [""]
    rt2_gutter_lines = [""]
    for range in data: 
        rt0_gutter_lines += ["", "", range.format_size(), "", "", ""]
        rt1_gutter_lines += ["", "", range.format_access(), "", "", ""]
        rt2_gutter_lines += ["", "", range.format_cacheable(), "", "", ""]
    rt0_gutter = "\n".join(rt0_gutter_lines)
    rt1_gutter = "\n".join(rt1_gutter_lines)
    rt2_gutter = "\n".join(rt2_gutter_lines)

    grid = Table.grid(padding=padding)
    grid.add_column(justify="right", width=left_width, no_wrap=True)
    grid.add_column(no_wrap=True)
    grid.add_column(justify="center", width=right0_width, no_wrap=True)
    if want_access_cacheable:
        grid.add_column(justify="center", width=right1_width, no_wrap=True)
        grid.add_column(justify="center", width=right2_width, no_wrap=True)
        grid.add_row("", "", column_titles[0], column_titles[1], column_titles[2])
        grid.add_row("", "", '-'*right0_width, '-'*right1_width, '-'*right2_width)
        grid.add_row(left, inner, rt0_gutter, rt1_gutter, rt2_gutter)
    else:
        grid.add_row("", "", column_titles[0])
        grid.add_row("", "", '-'*right0_width)
        grid.add_row(left, inner, rt0_gutter)
    return grid

def make_tower(slaves) -> str:
    """Make a tower visualization of the memory map"""

    console = Console()
    sorted_ranges = _create_ranges(slaves)
    # console.print(f"sorted_ranges: {sorted_ranges}")

    tower = stack(sorted_ranges, want_access_cacheable=True)
    with console.capture() as capture:
        console.print(tower)
    ascii = capture.get()
    return ascii



#!/usr/bin/env python3

import argparse
from dataclasses import dataclass
import os
import sys
import json
import shlex
from rich.text import Text
from rich import print, reconfigure, get_console

@dataclass
class SubInvocationItem:
    path: str
    args: argparse.Namespace

class SubInvocationItemList(list[SubInvocationItem]):
    def __init__(self, args: argparse.Namespace, json_data: list[dict]):
        super().__init__()
        self.args = args
        for item in json_data:
            path = item['path']
            args = item['args']
            args_arr:list[str] = []
            if args is not None:
                try:
                    if isinstance(args, str):
                        args_arr = shlex.split(args, posix=True)
                    else:
                        args_arr = args
                except ValueError:
                    args_arr = args
            # deep copy sys.argv[1:]
            combined_argv = sys.argv[1:].copy()
            combined_argv.extend(args_arr)
            combined_args = parse_args(combined_argv, override_paths=[path])
            self.append(SubInvocationItem(path=path, args=combined_args))

def parse_args(argv: list[str] = sys.argv[1:], override_paths: list[str] = None) -> argparse.Namespace:    
    class ColorReconfigureAction(argparse.BooleanOptionalAction):
        """
        BooleanOptionalAction that automatically reconfigures the default rich console to 
        disable colors if the "--no-colors" flag is used.  The option is then swallowed
        so it does not appear in the args Namespace after `args = parse_args(...)`
        """
        def __init__(self, option_strings, dest, **kwargs):
            # never create an attribute on the Namespace unless the flag appears,
            # and we’ll delete it anyway; default behavior is "colors on".
            kwargs.setdefault('default', argparse.SUPPRESS)
            super().__init__(option_strings, dest, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            # Treat '-C' as the negative form too.
            is_negative = (
                option_string in getattr(self, '_negative_option_strings', ())
            )

            if is_negative:
                # user asked to disable colors
                import os
                from rich import reconfigure
                os.environ["NO_COLOR"] = "1" # causes all calls to `c = Console()` to have colors off
                reconfigure(no_color=True)   # fallback to ensure default global rich console is replaced

            # swallow: remove the attribute so callers never see it
            vars(namespace).pop(self.dest, None)

    DEFAULT_EXTENSIONS = ['.sv', '.v', '.svh', '.vh']
    parser = argparse.ArgumentParser(description="Replace @subst...@endsubst blocks in a Verilog or SystemVerilog file")
    parser.add_argument("PATHS", nargs='+', type=str, help="File(s) or directory(ies) to process @subst blocks in; may be a single '@json:' prefixed filename with paths and args arrays inside it")
    file_opts = parser.add_argument_group("File Finding")
    file_opts.add_argument("--recursive", '-r', action="store_true", help="For directories in PATHS, whether to recurse into subdirectories")
    file_opts.add_argument("--extensions", '-e', nargs='+', default=DEFAULT_EXTENSIONS, type=str, help=f"File extensions to process within directories in PATHS (default: {', '.join(DEFAULT_EXTENSIONS)})")    
    control_opts = parser.add_argument_group("Control")
    control_opts.add_argument("--print-one-line", '-1', action="store_true", help="Print a single output line for each file processed (default: print nothing)")
    control_opts.add_argument("--force", '-f', action="store_true", help="Don't prompt for confirmation before overwriting a file (default: prompt for confirmation)")
    control_opts.add_argument("--colors", action=ColorReconfigureAction, help="Enable or disable colors in output (default: enabled)")
    control_opts.add_argument("--exit-code-if-modified", '-m', action="store_true", help="Exit with code 1 if any files were modified")
    debug_opts = parser.add_argument_group("Debugging")
    debug_opts.add_argument("--dry-run", action="store_true", help="Just show the diff, don't change any files")
    debug_opts.add_argument("--verbose", '-v', action="count", default=0, help="Show verbose output")
    debug_opts.add_argument("--debug-output-dir", '-o', action="store", default=None, help="Rather than in-place modifying the input file, write the output to same-named files in this directory")
    extra_args = parser.add_argument_group("Extra Arguments")
    extra_args.add_argument("--extra-args", '-x', action='append', nargs='*', type=str, help="Extra arguments to insert at the END of the script's argument list (can be specified multiple times)")
    extra_args.add_argument("--extra-args-leading", '-X', action='append', nargs='*', type=str, help="Extra arguments to insert at the BEGINNING of the script's argument list (can be specified multiple times)")
    args = parser.parse_args(argv)

    # if override_paths is not None, then we override the PATHS element with the given paths
    if override_paths is not None:
        args.PATHS = override_paths

    # Flatten the lists of lists that action='append' creates, or use empty list if none specified
    args.extra_args = [item for sublist in (args.extra_args or []) for item in sublist]
    args.extra_args_leading = [item for sublist in (args.extra_args_leading or []) for item in sublist]
    args.extensions = [ext if ext[0] == '.' else '.' + ext for ext in args.extensions]

    return args

def process_path(paths: list[str], extensions: list[str], recursive: bool, dry_run: bool, verbose: int, force: bool, extra_args: str, extra_args_leading: str, debug_output_dir: str|None, ret_paths_result: dict[str, bool] = dict[str, bool]()) -> None:
    """
    Processes a list of files or directories.

    Returns the paths_result dictionary.
    """
    for path in paths:
        if recursive and os.path.isdir(path):
            for root, dirs, _ in os.walk(path):
                for dir in dirs:
                    if verbose > 0:
                        print(f"Recursing into directory: {os.path.join(root, dir)}")
                    process_path([os.path.join(root, dir)], extensions, recursive, dry_run, verbose, force, extra_args, extra_args_leading, debug_output_dir, ret_paths_result)

        if os.path.isdir(path):
            if verbose > 0:
                print(f"Examining directory: {path}")
            for file in os.listdir(path):
                file_ext = os.path.splitext(file)[1]
                if verbose > 0:
                    print(f"Considering file: {os.path.join(path, file)} with extension '{file_ext}' against extensions: {extensions}")
                if any(file_ext == ext for ext in extensions):
                    if verbose > 0:
                        print(f"Recursing on file: {os.path.join(path, file)}")
                    process_path([os.path.join(path, file)], extensions, recursive, dry_run, verbose, force, extra_args, extra_args_leading, debug_output_dir, ret_paths_result)
                else:
                    if verbose > 0:
                        print(f"Skipping file: {os.path.join(path, file)}")
        else:
            if verbose > 0:
                print(f"Examining file: {path}")
            from .replace import run_substitution_on_file
            file_modified = run_substitution_on_file(file_path=path, dry_run=dry_run, verbose=verbose, force=force, extra_args=extra_args, extra_args_leading=extra_args_leading, debug_output_dir=debug_output_dir)
            ret_paths_result[path] = file_modified

def print_one_line_result(path: str, paths_result: dict[str, bool]) -> None:
    base_name = os.path.basename(path)
    if paths_result[path]:
        t = Text.assemble(
            "✔️ ",
                ("subst: " + base_name, "bold white"),
                (" (modified)", "bold yellow"),
            )
    else:
        t = Text.assemble(
            "✔️ ",
            ("subst: " + base_name + " (no change)", "dim"),
        )
    get_console().print(t)

def main():
    # Otherwise, normal operation
    args = parse_args()

    # collect the result of each path we process (True means file was modified, False means no change)
    ret_paths_result = dict[str, bool]()

    # If the PATHS element is prefixed with '@json:', then it is a JSON file with paths and args.
    # We re-invoke process_path for each element in the JSON array since each one
    # can have its own arguments
    if args.PATHS[0].startswith('@json:'):
        with open(args.PATHS[0][len('@json:'):], 'r') as f:
            json_data = json.load(f)
        args.PATHS = SubInvocationItemList(args, json_data)
        if isinstance(args.PATHS, SubInvocationItemList):
            for subinvocation_item in args.PATHS:
                process_path(paths=[subinvocation_item.path], 
                            extensions=args.extensions, 
                            recursive=args.recursive, 
                            dry_run=args.dry_run, 
                            verbose=args.verbose, 
                            force=args.force, 
                            extra_args=subinvocation_item.args.extra_args, 
                            extra_args_leading=subinvocation_item.args.extra_args_leading,
                            ret_paths_result=ret_paths_result,
                            debug_output_dir=args.debug_output_dir)
                print_one_line_result(subinvocation_item.path, ret_paths_result) if args.print_one_line else None
    else:
        for path in args.PATHS:
            process_path(paths=[path], 
                        extensions=args.extensions, 
                        recursive=args.recursive, 
                        dry_run=args.dry_run, 
                        verbose=args.verbose, 
                        force=args.force, 
                        extra_args=args.extra_args if args.extra_args is not None else "", 
                        extra_args_leading=args.extra_args_leading if args.extra_args_leading is not None else "", 
                        ret_paths_result=ret_paths_result,
                        debug_output_dir=args.debug_output_dir)
            print_one_line_result(path, ret_paths_result) if args.print_one_line else None

    if args.exit_code_if_modified and any(ret_paths_result.values()):
        sys.exit(1)
    else:
        sys.exit(0)

if __name__ == "__main__":
    main()